/***************************************************************************
 *   Copyright (C) 2007 by Julien Keller                                   *
 *   jkeller@imperial.ac.uk                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Curvatures.h"


#define pow2(x) (x*x)
#define pow3(x) (x*x*x)
#define pow4(x) pow2(x*x)
#define pow5(x) (x*pow4(x))
#define pow6(x) pow2(pow3(x))
#define pow8(x) pow2(pow4(x))
#define _Complex_I std::complex<double>(0, 1)


void MMatrixD2fm(Complex y[],const Complex Psi,const int Degree,const int N,MMatrix<Complex> &ddf)
{           // makes a matrix
  int i,j;
  for(i=0; i<Degree; i++)
  {
    for(j=i; j<Degree; j++)
    {
      ddf[i][j] = D2f(i,j,y,Psi,Degree,N);
      if(i!=j)
      {
        ddf[j][i]=ddf[i][j];
      }
    }
  }
}


Complex Der2SA(Complex y[], vector<int> sA, int i, int j)
{                 // d_i of sA at Aff pt y
  if(sA[i]==0 || sA[j]==0) return 0;
  else
  {
    if(i==j && sA[i]==1) return 0;
    else
    {
      sA[i] -= 1; sA[j] -= 1;
      Complex p=1.0;
      for(int k=0; k<Degree; k++) p*=pow(y[k], sA[k]);
      return ( (i==j) ? (Complex)((sA[i]+2)*(sA[i]+1)) : (Complex)((sA[i]+1)*(sA[j]+1)) )* p;
    }
  }
}

void MMatrixFDx_ix_jy_d(const MMatrix<Complex> &df, const MMatrix<Complex> &ddf,MMatrix<Complex> &m)
{           // m(section)(i,j)
  //   Matrix m(Dimension,Dimension);
  int i,j;
  for(i=0; i<Dimension; i++)
    for(j=i; j<Dimension; j++)
    {
      m[i][j] = -( ddf[i][j] - (ddf[i][Dimension]*df[j][0]+ddf[j][Dimension]*df[i][0])/df[Dimension][0] + ddf[Dimension][Dimension]*df[i][0]*df[j][0]/pow2(df[Dimension][0]) )/df[Dimension][0];
      if(i!=j) m[j][i] = m[i][j];
    }
  //  return m;
}



void All_DerSA(Complex y[],const int K, MMatrix<Complex> &Deriv_Result,const vector<vector<int> > &sA,const int NS,const MMatrix<Complex> &Result_sections)
{
  int i,deriv,k,j;


  Deriv_Result.fillWith(1);
  for(i=0;i<NS;i++)
  {
    for(deriv=0;deriv<Degree;deriv++)
    {
      if(sA[i][deriv]==0)
      {
        Deriv_Result[i][deriv]=0;
      }
      else
      {

        for(k=0;k<Degree;k++)
        {
          if(k==deriv)
          {
            Deriv_Result[i][deriv]*=Result_sections[k][sA[i][k]-1]*((Complex)(sA[i][k]));
          }
          else
          {
            Deriv_Result[i][deriv]*=Result_sections[k][sA[i][k]];

          }
        }
      }
    }
  }
}
void MMatrixFDy_psAbis(Complex y[], const vector<vector<int> > &sA, int NS,MMatrix <Complex> &v2, const int K,const MMatrix<Complex> &Result_sections)
{
  int i,j;
  MMatrix<Complex> Deriv_Result(NS,Degree);
  All_DerSA(y,K,Deriv_Result,sA,NS,Result_sections);
  for(i=0; i<Degree; i++)
    for(j=0; j<NS; j++) v2[i][j] = Deriv_Result[j][i];
}




void MMatrixFDy_py_qsA(Complex y[], const vector<vector<int> > &sA,
                       MMatrix<MMatrix<Complex> > &w)
{                 // dy_p of sAlpha returns a D x NS matrix
  int i,j, NS=sA.size();
  // u::zero_matrix<Complex> ZeroD(Degree,Degree);
  for(int A=0; A<NS; A++)
  {
    w[A][0].resize(Degree,Degree);
    w[A][0].fillWith(0);
    for(i=0; i<Degree; i++)
      for(j=i; j<Degree; j++)
      {
        w[A][0][i][j] = Der2SA(y, sA[A], i,j);
        if(i!=j) w[A][0][j][i] = w[A][0][i][j];
      }
  }
}


void MMatrixFDx_ix_jsA(Complex y[], const MMatrix<Complex> &df, const MMatrix<Complex> &ddf, const vector<vector<int> > &sA,MMatrix<MMatrix<Complex> > &DxixjsA,const MMatrix<Complex> &Result_sections,const int K)
{                 // m(section)(i,j)
  int i,j, p,q, NS=sA.size();
  MMatrix<Complex> v(Dimension,Degree); MMatrixFDx_iy_p(df,v);
  MMatrix<Complex> m(Dimension,Dimension); MMatrixFDx_ix_jy_d(df, ddf,m);
  MMatrix<MMatrix<Complex> > w(NS,1); MMatrixFDy_py_qsA(y, sA,w);
  //u::zero_matrix<Complex> Zero(Dimension,Dimension);
  MMatrix<Complex> v2(Degree,NS);
  MMatrixFDy_psAbis(y,sA,NS,v2,K,Result_sections);
  for(int A=0; A<NS; A++)
  {
    DxixjsA[A][0].resize(Dimension,Dimension);
    DxixjsA[A][0].fillWith(0);
    for(i=0; i<Dimension; i++)
      for(j=i; j<Dimension; j++)
      {
        for(p=0; p<Degree; p++)
          for(q=0; q<Degree; q++)
            DxixjsA[A][0][i][j] += v[i][p]*v[j][q]*w[A][0][p][q];
        if(i!=j) DxixjsA[A][0][j][i] = DxixjsA[A][0][i][j];
      }
    DxixjsA[A][0] += v2[Dimension][A]*m;
  }
}


void MMatrixD2_xP(const MMatrix<Complex> &G, const MMatrix<MMatrix<Complex> > &DxixjsA, const MMatrix<Complex> &sAy, int NS, MMatrix<Complex> &D2xP)
{
  //Matrix M(Dimension,Dimension); M.clear();
  //D2xP.clear();
  int i,j,m,n;
  D2xP.fillWith(0);
  for(n=0; n<Dimension; n++)
    for(m=n; m<Dimension; m++)
    {
      for(i=0; i<NS; i++)
        for(j=0; j<NS; j++)
          D2xP[m][n] += G[i][j]*DxixjsA[i][0][m][n]*conj(sAy[j][0]);
      if(n!=m) D2xP[n][m] = D2xP[m][n];
    }
  // return M;
}


void MMatrixDDbD3_xP(const MMatrix<Complex> &G, const MMatrix<MMatrix<Complex> > &DxixjsA, const MMatrix<Complex> &DxisA, int NS,MMatrix<MMatrix<Complex> > &DDbD3xP)
{           // (a-holo)(holo,holo)
  //VectorM M(Dimension);
  int i,j,m,n,p;
  //u::zero_matrix<Complex> ZeroD(Dimension,Dimension);
  //DDbD3xP.fillWith(0);
  for(p=0; p<Dimension; p++)
  {
    DDbD3xP[p][0].resize(Dimension,Dimension);
    DDbD3xP[p][0].fillWith(0);
    for(n=0; n<Dimension; n++)
      for(m=0; m<Dimension; m++)
        for(i=0; i<NS; i++)
          for(j=0; j<NS; j++)
            DDbD3xP[p][0][m][n] += G[i][j]*DxixjsA[i][0][n][m]*conj(DxisA[j][p]);
  }
  //   return M;
}


void MMatrixDDbDD4_xP(const MMatrix<Complex> &G, const MMatrix<MMatrix<Complex> > &DxixjsA, int NS,MMatrix<MMatrix<Complex> > &D4xP)
{           // (a-holo,a-holo)(holo,holo)
  //   MatrixM M(Dimension,Dimension);
  //   u::zero_matrix<Complex> ZeroD(Dimension,Dimension);
  int i,j,m,n,p,q;
  // D4xP.fillWith(0);
  //   D4xP.resize(Dimension,Dimension);
  for(p=0; p<Dimension; p++)
    for(q=0; q<Dimension; q++)
    {
      D4xP[p][q].resize(Dimension,Dimension);
      D4xP[p][q].fillWith(0);
      for(n=0; n<Dimension; n++)
        for(m=0; m<Dimension; m++)
          for(i=0; i<NS; i++)
            for(j=0; j<NS; j++)
              D4xP[p][q][m][n] =D4xP[p][q][m][n]+G[i][j]*DxixjsA[i][0][m][n]*conj(DxixjsA[j][0][p][q]);
    }
  //   return M;
}



void MMatrixD1_xg(const Complex P,const MMatrix<Complex> &DxP,const MMatrix<Complex> &D2xP,const MMatrix<Complex> &DbD2xP,const MMatrix< MMatrix<Complex> > &DDbD3xP,  MMatrix<MMatrix <Complex> > &D1xg)
{     // first derivative of Ricci metric g
  int k,i,j;
  //   VectorM M(Dimension);
  //u::zero_matrix<Complex> ZeroD(Dimension,Dimension);
  for(k=0; k<Dimension; k++)
  {
    D1xg[k][0].resize(Dimension,Dimension);
    D1xg[k][0].fillWith(0);
    for(j=0; j<Dimension; j++)
      for(i=0; i<Dimension; i++)
        D1xg[k][0][i][j] =
          2.0/pow3(P)* DxP[k][0]*DxP[i][0]*conj(DxP[j][0])
          -1.0/pow2(P)*(D2xP[k][i]*conj(DxP[j][0]) + DxP[i][0]*DbD2xP[j][k] + DxP[k][0]*DbD2xP[j][i])+1.0/P*DDbD3xP[j][0][i][k];
  }
  //   return M;
}




void MMatrixDD2_xg(Complex P, const MMatrix<Complex> &DxP,const MMatrix<Complex> &D2xP,const MMatrix<Complex> &DbD2xP,const MMatrix< MMatrix<Complex> > &DDbD3xP,const MMatrix< MMatrix<Complex> > &D4xP,MMatrix< MMatrix<Complex> > &DD2xg)
{            // second derivative of Ricci metric g
  int m,i,j,k;
  DD2xg.fillWith(0);
  // MatrixM M(Dimension,Dimension);
  for(m=0; m<Dimension; m++)
    for(k=0; k<Dimension; k++)
    {
      DD2xg[k][m].resize(Dimension,Dimension);
      //DD2xg[k][m].fillWith(0);
      for(i=0; i<Dimension; i++)
      {
        for(j=0; j<Dimension; j++)
        {
          DD2xg[k][m][i][j] =
            -6.0/pow4(P) * conj(DxP[m][0])*conj(DxP[j][0])*DxP[k][0]*DxP[i][0] +
            2.0/pow3(P) * ( DbD2xP[m][k]*DxP[i][0]*conj(DxP[j][0]) + DxP[k][0] * ( DbD2xP[m][i]*conj(DxP[j][0]) + conj(D2xP[m][j])*DxP[i][0] )
                            + conj(DxP[m][0]) * ( D2xP[k][i]*conj(DxP[j][0]) + DxP[i][0]*DbD2xP[j][k] + DxP[k][0]*DbD2xP[j][i] )  )
            -1.0/pow2(P) * ( conj(DxP[m][0])*DDbD3xP[j][0][i][k] + DbD2xP[m][k]*DbD2xP[j][i] + DxP[k][0]*conj(DDbD3xP[i][0][m][j]) +
                             conj(DxP[j][0])*DDbD3xP[m][0][k][i]+ D2xP[k][i]*conj(D2xP[m][j]) + DbD2xP[m][i]*DbD2xP[j][k] +
                             DxP[i][0]*conj(DDbD3xP[k][0][m][j])    )
            +1.0/P * D4xP[m][j][i][k];
        }
      }
    }
  //   return M;
}
/*
void MMatrixDD2_xgbis(Complex P, const MMatrix<Complex> &DxP,const MMatrix<Complex> &D2xP,const MMatrix<Complex> &DbD2xP,const MMatrix< MMatrix<Complex> > &DDbD3xP,const MMatrix< MMatrix<Complex> > &D4xP,MMatrix <MMatrix<Complex> > &DD2xg)
{            // second derivative of Ricci metric g
   int m,i,j,k;
   // MatrixM M(Dimension,Dimension);
   u::zero_matrix<Complex> ZeroD(Dimension,Dimension);
   for(m=0; m<Dimension; m++)
      for(k=0; k<Dimension; k++)
      {
         DD2xg(k,m)=1.0*ZeroD;
         for(i=0; i<Dimension; i++)
            for(j=0; j<Dimension; j++)
               DD2xg(k,m)(i,j) =
                  -6.0/pow4(P) * conj(DxP[m][0])*conj(DxP[j][0])*DxP[k][0]*DxP[i][0] +
                  2.0/pow3(P) * ( DbD2xP[m][k]*DxP[i][0]*conj(DxP[j][0]) + DxP[k][0] * ( DbD2xP[m][i]*conj(DxP[j][0]) + conj(D2xP[m][j])*DxP[i][0] )
                                  + conj(DxP[m][0]) * ( D2xP[k][i]*conj(DxP[j][0]) + DxP[i][0]*DbD2xP[j][k] + DxP[k][0]*DbD2xP[j][i] )  )
                  -1.0/pow2(P) * ( conj(DxP[m][0])*DDbD3xP[j][0][i][k] + DbD2xP[m][k]*DbD2xP[j][i] + DxP[k][0]*conj(DDbD3xP[i][0][m][j]) +
                                   conj(DxP[j][0])*DDbD3xP[m][0][k][i]+ D2xP[k][i]*conj(D2xP[m][j]) + DbD2xP[m][i]*DbD2xP[j][k] +
                                   DxP[i][0]*conj(DDbD3xP[k][0][m][j])    )
                  +1.0/P * D4xP[m][j][i][k];
      }
   //   return M;
}
*/

/*
extern "C"
{
  //void zgetrf_(const int &N, const int &N, cx *, const int &N, int * ,       const int &N);
  //void zgetri_(const int &N,               cx *, const int &N, int * , cx *, const int &N, const int &N);
  void zhetrf_(const char &UPLO, const int &N, Complex *, const int &N, int *, Complex *, const int &LWORK, const int &INFO);
  void zhetri_(const char &UPLO, const int &N, Complex *, const int &N, int *, Complex *, const int &INFO);
}


void MMatrixInvertF(const MMatrix<Complex> &M, MMatrix<Complex> &Mi, double ERR)
{
  int i,j, N=M.getHeight();
  int INFO, IPIV[N];
  Complex  WORK[N], *A = new Complex [N*N];
  for(i=0; i<N; i++)
    for(j=i; j<N; j++)
      A[i*N+j] = M[i][j];

  zhetrf_('L',N, A, N, IPIV, WORK, N, INFO);
  zhetri_('L',N, A, N, IPIV, WORK,    INFO);

  for(i=0; i<N; i++)
  {
    Mi[i][i] = A[i*N+i];
    for(j=i+1; j<N; j++)
    {
      Mi[i][j] = A[i*N+j];
      Mi[j][i] = conj(Mi[i][j]);
    }
  }
  delete [] A;

}
*/


double Ricci_FAST(const double ERR,Complex P, MMatrix<Complex> &DxP,const int K,const MMatrix<Complex> &D2xP, const MMatrix<Complex> &DbD2xP,const MMatrix<MMatrix<Complex> > &DDbD3xP,const MMatrix<MMatrix<Complex> > &D4xP,const MMatrix<Complex> &g,MMatrix<Complex> &Ric)
{           // compute Ricci tensor need to compute g inverse, use inv(g) for the inverse of g
  int p,m,i,j,k,l;

  MMatrix< MMatrix<Complex> > D1xg(Dimension,1);


  /*   cout << " P="<<P  << endl << "DxP\n ";
     show(DxP); 
     cout << "\n D2xP et DbD2xP";
     show(D2xP);
     show(DbD2xP);
     cout << "DDbD3xP\n";
     for(int k0=0;k0<Dimension;k0++)
     {
         show(DDbD3xP[k0][0]);
     } */
  MMatrixD1_xg(P,DxP,D2xP,DbD2xP,DDbD3xP,D1xg);
  /*cout << "D1xg";
  for(int k0=0;k0<Dimension;k0++)
  {
      show(D1xg[k0][0]);
  }*/
  //Matrix Ric(Dimension,Dimension);
  //Ric.clear();
  MMatrix< MMatrix<Complex> > DD2xg(Dimension,Dimension);
  /*cout << "D4xP";
  for(int k0=0;k0<Dimension;k0++)
  {
   for(int k1=0;k1<Dimension;k1++){
         show(D4xP[k0][k1]);
  }
  } */

  MMatrixDD2_xg(P,DxP,D2xP,DbD2xP,DDbD3xP,D4xP,DD2xg);

  // cout << "\n DD2xg[0][0]";
  // show(DD2xg[0][0]);
  MMatrix<Complex> gi(Dimension,Dimension);
//  MMatrixInvertF(g, gi, ERR);///TODO variable precision
  Invert2( g, gi, ERR);
  //show(gi);
  Ric.fillWith(0);
  for(i=0; i<Dimension; i++)
    for(j=i; j<Dimension; j++)
    {
      for(p=0; p<Dimension; p++)
        for(k=0; k<Dimension; k++)
        {
          Ric[i][j] += gi[k][p] * DD2xg[i][j][p][k];
          for(m=0; m<Dimension; m++)
            for(l=0; l<Dimension; l++)
              Ric[i][j] -= gi[k][p] * D1xg[i][0][p][m]*gi[m][l]*conj(D1xg[j][0][k][l]);
        }
      if(i!=j) Ric[j][i] = conj(Ric[i][j]);
    }

  Complex trace=0;
  for(i=0; i<Dimension; i++)
    for(j=0; j<Dimension; j++)
      trace += gi[i][j]*Ric[j][i];
  trace *= Complex(K);
  //   if(imag(trace) > 100*ERR) cout << "ERROR = " << imag(trace) << '\n';
  ///should disappear
  return real(trace);

}


template<class T> void writeG(ofstream &file, T x)
{
  for(int i=0; i<Dimension; i++)
    for(int j=0; j<Dimension; j++)
      file << x(i,j); file << '\n';
}

void MMatrixFDx_iy_p(const MMatrix<Complex> &df, MMatrix<Complex> &v_deriv)      // compute the matrix dx_i of y_p
{                             // y_p coords on C^D, x_i coords on X
  // Matrix m(Dimension,Degree);
  ///v.clear(); ??
  for(int i=0; i<Dimension; i++)
  {
    v_deriv[i][i] = 1.0;
    v_deriv[i][Dimension] = -df[i][0]/df[Dimension][0];
  }
  // return m;
}



void MMatrixFDx_isAbis(Complex y[], const vector<vector<int> > &sA, int NS,MMatrix<Complex> &DxisA,const int K,const MMatrix<Complex> &Result_sections, const MMatrix<Complex> &v_deriv)
{
  //  Matrix v_deriv(Dimension,Degree);
  //  MMatrix<Complex> DxisAtemp(NS,Dimension);
  MMatrix<Complex> v2(Degree,NS);
  MMatrixFDy_psAbis(y,sA,NS,v2,K,Result_sections);
  // cout << v2 <<endl;

  // All_DerSA(y,K,v2,sA,NS,Result_sections);
  // cout << v2 <<endl;
  //m= prod(v, v2);
  int i,j,k;
  DxisA.fillWith(0);
  for(i=0;i<NS;i++)
  {
    for(j=0;j<Dimension;j++)
    {
      //      DxisA[i][j]=DxisAtemp(i,j);
      for(k=0;k<Degree;k++)
      {
        DxisA[i][j]+=v2[k][i]*v_deriv[j][k];
      }
    }
  }

}


void Computations_of_sections(Complex y[],const int K,MMatrix<Complex> &Result_sections)
{
  int i,j;
  Result_sections.fillWith(1);
  for(i=0;i<Degree;i++)
  {
    for(j=1;j<K+1;j++)
    {
      Result_sections[i][j]=Result_sections[i][j-1]*y[i];
    }
  }

}

void MMatrixsA_ybis(const vector<vector<int> > &sA,const int NS,MMatrix<Complex> &sAy,const MMatrix<Complex> &Result_sections)
{           // evaluates sA at point y on X
  //Vector m(NS);
  int i,j;
  for( i=0; i<NS; i++)
  {
    sAy[i][0]=1.0;
    for( j=0; j<Degree; j++) sAy[i][0] *= Result_sections[j][sA[i][j]];
  }
  //return m;


}

void VVectorsA_ybis(const vector<vector<int> > &sA,const int NS,VVector<Complex> &sAy,const MMatrix<Complex> &Result_sections)
{           // evaluates sA at point y on X
  //Vector m(NS);
  for(int i=0; i<NS; i++)
  {
    sAy[i]=1.0;
    for(int j=0; j<Degree; j++) sAy[i] *= Result_sections[j][sA[i][j]];
  }
  //return m;


}




Complex MMatrixFPbis(const int NS,const MMatrix<Complex> &sAy,const MMatrix<Complex> &G) // compute P at point y
{
  Complex s=0;
  int i,j;
  for(i=0; i<NS; i++)
    for(j=i+1; j<NS; j++)
      s += 2*real(G[i][j]*sAy[i][0]*conj(sAy[j][0]));
  for(i=0;i<NS;i++)
  {
    s+=G[i][i]*pow2(abs(sAy[i][0]));
  }
  return s;
}


vector<vector<int> > sAaff(vector<vector<int> > sA, int k)
{           // erase the component that became affine
  int i,j;
  vector<int>::iterator J;
  for (int i=0; i<sA.size(); i++)
  {
    J=sA[i].begin();
    for(int j=0; j<k; j++) J++;
    sA[i].erase(J);
  }
  return sA;
}

void MMatrixD_xP(const int NS,const MMatrix<Complex> &DxisA,const MMatrix<Complex> &G,const MMatrix<Complex> &sAy, MMatrix<Complex> &DxP)
{  // Compute various derivatives of P
  //Vector v(Dimension);
  int i,j,m;
  DxP.fillWith(0);
  for(m=0; m<Dimension; m++)
    for(i=0; i<NS; i++)
    {
      for(j=0; j<NS; j++)
      {
        DxP[m][0] += ((Complex) G[i][j])*
                     DxisA[i][m]*conj(sAy[j][0]);

      }

    }
  //  return v;
}


void MMatrixDbD2_xPbis(const int NS,const MMatrix<Complex> &DxisA,const MMatrix<Complex> &G,MMatrix<Complex> &DbD2xP)
{           // (a-holo,holo)
  //    Matrix M(Dimension,Dimension);
  //u::zero_matrix<Complex> Zero(Dimension,Dimension); //clear ?
  //    M.clear();
  DbD2xP.fillWith(0);
  int i,j,m,n;
  for(n=0; n<Dimension; n++)
  {
    for(m=n; m<Dimension; m++)
    {
      if(n!=m)
      {
        for(i=0; i<NS; i++)
        {
          for(j=0;j<NS; j++)
          {
            DbD2xP[m][n] += G[i][j]*DxisA[i][n]*conj(DxisA[j][m]);
          }
        }

        DbD2xP[n][m]=conj(DbD2xP[m][n]);
      }
      else
      {
        for(i=0; i<NS; i++)
        {
          for(j=i+1;j<NS; j++)
          {

            DbD2xP[m][n] += 2*real(G[i][j]*DxisA[i][n]*conj(DxisA[j][m]));

          }
          DbD2xP[m][m] += G[i][i]*pow2(abs(DxisA[i][m]));
        }

      }

    }
    // return M;
  }
}




void MMatrixFgbis(const MMatrix<Complex> &DxP,MMatrix<Complex> &DbD2xP,const Complex P,MMatrix<Complex> &g)
{   // Ricci metric g
  int i,j;
  //Matrix M(Dimension,Dimension);
  for(i=0; i<Dimension; i++)
    for(j=i; j<Dimension; j++)
    {
      g[i][j] = (1.0/P)*DbD2xP[j][i] - (1.0/pow2(P))*DxP[i][0]*conj(DxP[j][0]);
      if(i!=j) g[j][i] = conj(g[i][j]);
    }
  //  cout<<" Test";
  //  show(g);
  //  getchar();
}

/// ------------- ADDED ---------------END ///
/*
cx getVol(const u::vector<Point> &vX, const vector<vector<int> > &Sec)
{
  int k;
  cx vol=0, y[D];
  Vector df(D); Matrix ddf(D,D);
  vector<vector<int> > Se;
  Point X;
 
  for(int i=0; i<vX.size(); i++)
{
    X = vX[i];
    k = ChooseMax(X);
    AffineCoord(X, y, k);
    Se = sAaff(Sec, k);
    df = ChooseY4(y); //cout<<abs(f(y))<<' ';
    ddf = D2fm(y);
    DxisA = FDx_isA(y,df, Se, NS);
    sAy = sA_y(y, Se);
    P = FP();
    DxP = D_xP();
    DbD2xP = DbD2_xP();
    g = Fg();//printM(g);
    vol += getOneVol(X, g);
}
  return vol/cx(vX.size());
}
 
VectorD DonaldsonEta(const u::vector<Point> &vX, const Matrix &Gi, const vector<vector<int> > &Sec)
{ // Eta is w_B^{3}/\Omega\wedge\bar{\Omega}
  int k,i, NP=vX.size();
  cx y[D];
  double tmp=0;
  Vector df(D); Matrix ddf(D,D);
  vector<vector<int> > Se;
  Point X;
  VectorD v(NP);
 
  for(i=0; i<NP; i++)
{
    X = vX[i];
    k = ChooseMax(X);
    AffineCoord(X, y, k);
    Se = sAaff(Sec, k);
    df = ChooseY4(y); //cout<<abs(f(y))<<' ';
    ddf = D2fm(y);
    DxisA = FDx_isA(y,df, Se, NS);
    sAy = sA_y(y, Se);
    P = FP();
    DxP = D_xP();
    DbD2xP = DbD2_xP();
    g = Fg();//printM(g);
    v(i) = real( getOneVol(X, g) / X.mass );
}
  for(i=0; i<NP; i++) tmp += v(i);  tmp /= NP;
  for(i=0; i<NP; i++) {v(i) /= tmp;cout<<v(i)<<' ';}
  return v;
}
 
////////// Sergio writes beautiful code
 
cx fAB(const Matrix &Gi, int A, int B, const Matrix &sAB)
{
  cx f;
 
  cx denom = 0;
  int a,b;
  for(a=0; a<NS; a++)
    for(b=0; b<NS; b++)
      denom += Gi(a,b) * sAB(a,b);
  f = sAB(A,B) / denom;
  return f;
}
 
Matrix etaMat(const u::vector<Point> &X, const Matrix &Gi, const VectorM &sAB, const vector<vector<int> > &s, VectorD &eta)
{
  Matrix etaM(NS,NS);
  cx t;
  int k,a,b, NP=X.size();
  eta = DonaldsonEta(X,Gi,s);//printV(eta);
 
  for(a=0; a<NS; a++)
    for(b=a; b<NS; b++)
{
      t=0;
      for(k=0; k<X.size(); k++)
        t += X(k).mass * (eta(k)-1) * fAB(Gi, a,b, sAB(k));
      etaM(a,b) = t;
      if(a!=b) etaM(b,a) = conj(t);
}
  return etaM;
}
*/

///////////////////////////////////////////////////////



double VolumeForm_FSmetric2_FAST2(const Point &Pt, ResultsStore &additResults,MMatrix<Complex> &G,MMatrix<Complex> &df,MMatrix<Complex> &ddf,MMatrix<Complex> &DxisA,vector<vector<int> > &Se,MMatrix<Complex> &sAy,MMatrix<Complex> &DxP,MMatrix<Complex> &DbD2xP,MMatrix<Complex> &g,MMatrix<Complex> &FS,MMatrix<Complex> &O,MMatrix<Complex> &v_deriv,const Complex Psi)
{
  int i,j;
  int NS=additResults.getNS();
  int K=additResults.getK();

  // MatrixD G(NS,NS); //this contains the hermitian matrice
  //cout << "\n G== "<< Temp.getG();
  // MatrixD G2(NS,NS);
  /*  cout << additResults.getvi() << endl;
  cout << additResults.getBs()<<endl;
  cout <<  additResults.getBl()<<endl;
  print(additResults.getPairs()); 
  */


  PutToFullMatrix(additResults.getvi(),G, additResults.getPairs(), additResults.getBs(), additResults.getBl());

  //cout<< G2 <<endl<<endl;
  //Temp.storeG(G2);
  //cout << "\n G== "<< Temp.getG();
  //cout << G <<endl;
  Complex y[Degree];
  /*  Vector df(Degree); Matrix ddf(Degree,Degree);
  Matrix DxisA(NS,Dimension);
  vector<vector<int> > Se;
  Vector sAy(NS);
  Vector DxP(Dimension), cDxP(Dimension);
  Matrix DbD2xP(Dimension,Dimension),g(Dimension,Dimension);
  */
  int k = ChooseMax(Pt);
  AffineCoord(Pt, y, k);
  Se = sAaff(additResults.getSec(), k);
  df = MMatrixChooseY4(y,Psi); // cout<<abs(f(y))<<' ';
  MMatrixFDx_iy_p(df,v_deriv);

  MMatrixD2fm(y,Psi,Degree,N,ddf);

  MMatrix<Complex> Result_sections(Degree,K+1);

  Computations_of_sections(y,K,Result_sections);
  MMatrixsA_ybis( Se,NS,sAy,Result_sections);

  Complex P = MMatrixFPbis(NS,sAy,G);

  MMatrixFDx_isAbis(y, Se, NS,DxisA,K,Result_sections,v_deriv);

  MMatrixD_xP(NS,DxisA,G,sAy,DxP);//
  MMatrixDbD2_xPbis(NS,DxisA,G,DbD2xP);
  MMatrixFgbis(DxP,DbD2xP,P,g);


  MMatrixFSmetric(y,FS);

  for (i=0; i<Dimension; i++)
    for (j=i; j<Dimension; j++)
    {
      O[i][j] = FS[i][j] + FS[Dimension][j]*(v_deriv[i][Dimension]) + (FS[i][Dimension] + FS[Dimension][Dimension]*(v_deriv[i][Dimension])) *  conj(v_deriv[j][Dimension]);
      if(i!=j) O[j][i] = conj(O[i][j]);
    }
  //  cout <<"\n O"<< O << endl;

  Complex ratio = determinant(g)/determinant(O);
  if(real(ratio)<0)
  {
    cout << "\n  <!>  det<0 : "<< real(determinant(g));
    getchar();
  }
  return(real(ratio));
  //
  // return(real( determinant(g) ));

}
double Find_Ricci_FAST(const double ERR,const Point Pt,const MMatrix<Complex> &G,const int NS,const int K,const vector<vector<int> > &Sec,const Complex Psi,MMatrix<Complex> &g, MMatrix<Complex> &df,MMatrix<Complex> &ddf, int k,Complex y[],MMatrix<Complex> &DxisA,MMatrix<MMatrix<Complex> > &DxixjsA,vector<vector <int> > &Se,MMatrix<Complex> &sAy, MMatrix<Complex> &DxP, MMatrix<Complex> &DbD2xP,MMatrix<Complex> &D2xP,MMatrix<Complex> &Ric,MMatrix<MMatrix<Complex> > &DDbD3xP,MMatrix<MMatrix<Complex> > &D4xP,MMatrix<Complex> &v_deriv)
{

  k = ChooseMax(Pt);

  AffineCoord(Pt, y, k);
  Se = sAaff(Sec, k);
  df = MMatrixChooseY4(y,Psi); // cout<<abs(f(y))<<' ';
  MMatrixFDx_iy_p(df,v_deriv);

  MMatrixD2fm(y,Psi,Degree,N,ddf);
  MMatrix<Complex> Result_sections(Degree,K+1);
  Computations_of_sections(y,K,Result_sections);
  MMatrixsA_ybis(Se,NS,sAy,Result_sections);


  Complex P = MMatrixFPbis(NS,sAy,G);

  //P = FP(NS,sAy,G);


  MMatrixFDx_isAbis(y, Se, NS,DxisA,K,Result_sections,v_deriv);

  MMatrixD_xP(NS,DxisA,G,sAy,DxP);//
  MMatrixFDx_ix_jsA(y,df,ddf, Se,DxixjsA,Result_sections,K);///
  MMatrixD2_xP(G,DxixjsA,sAy,NS,D2xP);///

  MMatrixDbD2_xPbis(NS,DxisA,G,DbD2xP);

  MMatrixFgbis(DxP,DbD2xP,P,g);
  MMatrixDDbD3_xP(G,DxixjsA,DxisA,NS,DDbD3xP);
  MMatrixDDbDD4_xP(G,DxixjsA,NS,D4xP);

  return Ricci_FAST(ERR,P,DxP,K,D2xP,DbD2xP,DDbD3xP,D4xP,g,Ric);

}





void Global_Find_Ricci_FAST(const double ERR,vector<Point> &X,const MMatrix<Complex> &G,const int NS,const int K,const vector<vector<int> > Sec,const Complex Psi,const int Number_Points)
{
  int i;

  double value_ricci1,value_ricci2;
  double st=0;
  double  tm=0;   /// Invert<double>(G,Gi,0.001);
  MMatrix<Complex> df(Degree,1);
  MMatrix<Complex> ddf(Degree,Degree);
  int k;
  MMatrix<Complex> v_deriv(Dimension,Degree);
  Complex y[Degree];
  MMatrix<Complex> DxisA(NS,Dimension);
  MMatrix<MMatrix<Complex> > DxixjsA(NS,1);
  vector<vector<int> > Se;
  MMatrix<Complex> sAy(NS,1);
  MMatrix<Complex> DxP(Dimension,1);
  MMatrix<Complex> DbD2xP(Dimension,Dimension),D2xP(Dimension,Dimension),Ric(Dimension,Dimension),g(Dimension,Dimension);
  MMatrix<MMatrix<Complex> > DDbD3xP(Dimension,1);
  MMatrix<MMatrix<Complex> > D4xP(Dimension,Dimension);
  //MMatrix<double> results_scal(50,1);
  //    Test
  /*  X[0].comp[0]=-0.559599784177818+0.0609561990154176*_Complex_I;
    X[0].comp[1]=0.0638399446822967+0.474411652715583*_Complex_I;
    X[0].comp[2]=-0.121771128485202-0.213496713157267*_Complex_I;
    X[0].comp[3]=0.0562870662879273-0.271454482868225*_Complex_I;
    X[0].comp[4]=-0.374146436818869+0.420404671755718*_Complex_I;
    X[0].mass=1.73751800254407;
    /*g[3,3](((2.15,1.11e-15),(0.133,-0.148),(0.212,-0.117)),((0.133,0.148),(1.49,-8.33e-17),(-0.0431,-0.0316)),((0.212,0.117),(-0.0431,0.0316),(1.59,-2.78e-17)))
     Scal = -1.54

    */
  double mass_total=0;
  for(i=0; i<Number_Points; i++)
  {
    mass_total+=X[i].mass;
  }
  mass_total=mass_total/Number_Points;  //TODO :has already been done, could be removed
  double results_norm_scal=0;
  double max_abs_scalar=0;
  double max_abs_scalar_weighted=0;

  for(i=0; i<(Number_Points)/2; i++)
  {
    value_ricci1=Find_Ricci_FAST(ERR,X[i],G,NS,K,Sec,Psi,g,df,ddf,k,y,DxisA,DxixjsA,Se,sAy, DxP,  DbD2xP,D2xP,Ric,DDbD3xP,D4xP,v_deriv);
    results_norm_scal += abs(value_ricci1)*(X[i].mass);
    if (abs(value_ricci1)>max_abs_scalar)  max_abs_scalar=abs(value_ricci1);
    if (abs(value_ricci1)*X[i].mass/Number_Points>max_abs_scalar_weighted)  max_abs_scalar_weighted=abs(value_ricci1)*X[i].mass/Number_Points;
    //cout << "\n value ="<< value_ricci1 << " "  << abs(value_ricci1)*(X[i].mass/Number_Points) << " \n" ;
    //getchar();
    value_ricci2=Find_Ricci_FAST(ERR,X[Number_Points-i-1],G,NS,K,Sec,Psi,g,df,ddf,k,y,DxisA,DxixjsA,Se,sAy, DxP,  DbD2xP,D2xP,Ric,DDbD3xP,D4xP,v_deriv);
    results_norm_scal += abs(value_ricci2)*(X[Number_Points-i-1].mass);
    if (abs(value_ricci2)>max_abs_scalar)  max_abs_scalar=abs(value_ricci2);
    if (abs(value_ricci2)*X[Number_Points-i-1].mass/Number_Points>max_abs_scalar_weighted)  max_abs_scalar_weighted=abs(value_ricci2)*X[Number_Points-i-1].mass/Number_Points;
    st += (value_ricci1*X[i].mass+value_ricci2*X[Number_Points-i-1].mass);
    //
    //      mass_computed_points+=X[i].mass;

    //results_scal[i % 25][0]=value_ricci*pow(X[i].mass/mass_total,1./Dimension);
    //value_ricci/X[i].mass;
    if(i % 20 ==0 && i>0 )
    {
      cout << " -----------------> L^1 Norm scalar curvature = " << results_norm_scal/(2*i*mass_total) << " Average scal = "<< st/(2*i*mass_total)<< "  \n";

      //results_scal.norm2()/25 << "\n" ;
    }

    cout.flush();
    printf("%4dth point:Scal,Scal weighted +their max:%3.4f %3.4f | %3.3f  %3.3f",2*i,(value_ricci1+value_ricci2)/2.,2*(value_ricci1*X[i].mass+value_ricci2*X[Number_Points-i-1].mass)/Number_Points,max_abs_scalar,max_abs_scalar_weighted);
    cout <<"\r";
    //      cout<<i<<' ' <<value_ricci << " "<< value_ricci/X[i].mass  << " " << value_ricci*pow(X[i].mass/mass_total,1./Dimension)<< '\r';
    // getchar();
    //         V(i) = Ricci(X[i],DxP);cout<<i<<' ' << V(i)<<'\n';
    //st += 2*(value_ricci1*X[i].mass+value_ricci2*X[Number_Points-i-1].mass)/(Number_Points*mass_total);
    ////   tm += X[i].mass;
  }

  //// st /= tm;
  //cout<< "\n Average scalar curvature=  " << st << endl;



}
/*
   for(i=0; i<Number_Points; i++)
   {
      value_ricci=Find_Ricci_FAST(ERR,X[i],G,NS,K,Sec,Psi,g,df,ddf,k,y,DxisA,DxixjsA,Se,sAy, DxP,  DbD2xP,D2xP,Ric,DDbD3xP,D4xP,v_deriv);
      results_scal[i % 50][0]=value_ricci*pow(X[i].mass/mass_total,1./Dimension);
//value_ricci/X[i].mass;
      if(i % 50 ==0 && i>0 )
      {
         cout << " ---------------------> Average = " << results_scal.sum()/50 << "\n" ;
      } 
      cout<<i<<' ' <<value_ricci << " "<< value_ricci/X[i].mass  << " " << value_ricci*pow(X[i].mass/mass_total,1./Dimension)<< '\n';
     // getchar();
      //         V(i) = Ricci(X[i],DxP);cout<<i<<' ' << V(i)<<'\n';
      st += value_ricci*X[i].mass;
      tm += X[i].mass;
   }
   
   st /= tm; cout<< "Average scalar curvature= " << st/Number_Points <<endl;
 
 
 
}
 
*/








void Deriv_pointwise(Point &X,const vector<vector<int> > &Sec,const int K,MMatrix<Complex> &Result0,MMatrix<Complex> &Result1,MMatrix< MMatrix<Complex> >&Result2, MMatrix< MMatrix<Complex> >&Result3, MMatrix< MMatrix< MMatrix< Complex> > >&Result4, MMatrix<Complex> &ValuePoints)
//here Result1 if a vector of complex of size the number of sections NS.
// we derive one time wrt to the variable number = derivative
// and Result containts all i_0 x_1^{i_1}...x_{i0}^{i0-1}...
//Result2 (i,d1,d2)=Result2[i][1][d1][d2] contains for the section i, the derivative wrt dz/dz_{d1} dz/dz_{d2}

//Result3 (i,d1,d2,d3)=Result3[i][d1][d2][d3] contains for the section i, the derivative wrt dz/dz_{d1} dz/dz_{d2} dz/dz3

//Result4 (i,0,d1,d2,d3,d4)=Result4[i][0][d1][d2][d3][d4] contains for the section i, the derivative wrt dz/dz_{d1} dz/dz_{d2} dz/dz3 dz/dz4
//Result0 = contains no derivatives, just for evaluation standard of sections
{
  int i,j;
  int derivative,derivative1,derivative2,derivative3,derivative4;  // stands for d/dz_i where i=derivative
  int Npt=X.length();
  for(i=0;i<Npt;i++)
  {
    ValuePoints[i][0]=1.0;
  }
  for(i=0;i<Npt;i++)
  {
    ValuePoints[i][1]=X.comp[i];
  }
  for(i=0;i<Npt;i++)
  {
    for(j=2;j<K+1;j++)
    {
      ValuePoints[i][j]=ValuePoints[i][j-1]*ValuePoints[i][1];
    }
  }
  for(i=0;i<Sec.size();i++)
  {
    Result0[i][0]=1.0; //erase previous results
    for(j=0;j<Npt;j++)
    {
      Result0[i][0]=Result0[i][0]*ValuePoints[j][Sec[i][j]];
    }
  }
  //Result1.resize(Sec.size(),Npt);
  for(derivative=0;derivative<Npt;derivative++)
  {

    for(i=0;i<Sec.size();i++)
    {
      Result1[i][derivative]=1.0; //erase previous results
      for(j=0;j<Npt;j++)
      {
        if(j==derivative)
        {
          if(Sec[i][j]==0)
          {
            Result1[i][derivative]=0; //TODO one could be more efficient and avoid other multiplications here
          }
          else
          {
            Result1[i][derivative]*=((Complex ) Sec[i][j])*ValuePoints[j][Sec[i][j]-1];
          }
        }
        else
        {
          if(Result1[i][derivative] != ((Complex) 0))
          {
            Result1[i][derivative]*=ValuePoints[j][Sec[i][j]];
          }
        }
      }
    }
  }
  //  show(ValuePoints);
  //  show(Result0);
  //  show(Result1);
  for(i=0;i<Sec.size();i++)
  {
    Result2[i][0].resize(Npt,Npt);
    Result2[i][0].fillWith(1.0);
  }
  for(derivative1=0;derivative1<Npt;derivative1++)
  {
    for(derivative2=derivative1;derivative2<Npt;derivative2++)
    {
      for(i=0;i<Sec.size();i++)
      {
        for(j=0;j<Npt;j++)
        {
          if(j==derivative1 && j==derivative2)
          {
            if(Sec[i][j]==0 || Sec[i][j]==1)
            {
              Result2[i][0][derivative1][derivative2]=0;
            }
            else
            {
              Result2[i][0][derivative1][derivative2]*=((Complex)  ((Sec[i][j])*(Sec[i][j]-1)))*ValuePoints[j][Sec[i][j]-2];
            }
          }
          else
          {
            if(j==derivative1 || j==derivative2)
            {
              if(Sec[i][j]==0)
              {
                Result2[i][0][derivative1][derivative2]=0;
              }
              else
              {
                Result2[i][0][derivative1][derivative2]*=ValuePoints[j][Sec[i][j]-1]*((Complex) Sec[i][j]);
              }
            }
            else
            {
              if(Result2[i][0][derivative1][derivative2] !=( (Complex) 0))
              {
                Result2[i][0][derivative1][derivative2]*=ValuePoints[j][Sec[i][j]];

              }
            }

          }
        }
        if(derivative2>derivative1)
        {
          Result2[i][0][derivative2][derivative1]=Result2[i][0][derivative1][derivative2];
        }
      }

    }
  }

  /// 3rd order derivatives

  //initialisation
  /*
  for(i=0;i<Sec.size();i++)
  {
    for(j=0;j<Npt;j++)
    {
      Result3[i][j].resize(Npt,Npt);
      Result3[i][j].fillWith(1.0);
    }
  }
  int case_derivs;
  for(derivative1=0;derivative1<Npt;derivative1++)
  {
    for(derivative2=0;derivative2<Npt;derivative2++)
    {
      for(derivative3=0;derivative3<Npt;derivative3++)
      {
        for(i=0;i<Sec.size();i++)
        {
          for(j=0;j<Npt;j++)
          {
            if(Sec[i][j]==derivative1 && Sec[i][j]==derivative2 && Sec[i][j]==derivative3)
            {
              case_derivs=3;
            }
            else
            {
              if((Sec[i][j]==derivative1 && Sec[i][j]==derivative2) ||
                  (Sec[i][j]==derivative1 && Sec[i][j]==derivative3) ||
                  (Sec[i][j]==derivative3 && Sec[i][j]==derivative2))
              {
                case_derivs=2;
              }
              else
              {
                if(Sec[i][j]==derivative1 || Sec[i][j]==derivative3 ||
                    Sec[i][j]==derivative2)
                {
                  case_derivs=1;
                }
                else
                {
                  case_derivs=0;
                }
              }
            }
            if(Sec[i][j]<case_derivs)
            {
              case_derivs=0;
            }
            switch(case_derivs)
            {
            case 0:
              Result3[i][derivative1][derivative2][derivative3]*=ValuePoints[j][Sec[i][j]];
              break;
            case 1:
              Result3[i][derivative1][derivative2][derivative3]*=((Complex) (Sec[i][j]) )*ValuePoints[j][Sec[i][j]-1];
              break;
            case 2:
                Result3[i][derivative1][derivative2][derivative3]*=((Complex) ((Sec[i][j])*(Sec[i][j]-1)))*ValuePoints[j][Sec[i][j]-2];
              break;
            case 3:
                Result3[i][derivative1][derivative2][derivative3]*=((Complex) ((Sec[i][j])*(Sec[i][j]-1)*(Sec[i][j]-2)))*ValuePoints[j][Sec[i][j]-3];
              break;
            }

          }
        }
      }
    }
  }



  /// 4th order derivatives

  //initialisation
  int r,s;
  for(i=0;i<Sec.size();i++)
  {
    for(j=0;j<Npt;j++)
    {
      Result4[i][0].resize(Npt,Npt);
      for(r=0;r<Npt;r++)
      {
        for(s=0;r<Npt;r++)
        {
          Result4[i][0][r][s].resize(Npt,Npt);
          Result4[i][0][r][s].fillWith(1.0);
        }
      }
    }
  }

  for(derivative1=0;derivative1<Npt;derivative1++)
  {
    for(derivative2=0;derivative2<Npt;derivative2++)
    {
      for(derivative3=0;derivative3<Npt;derivative3++)
      {
        for(derivative4=0;derivative4<Npt;derivative4++)
        {
          for(i=0;i<Sec.size();i++)
          {
            for(j=0;j<Npt;j++)
            {
              if(Sec[i][j]==derivative1 && Sec[i][j]==derivative2 && Sec[i][j]==derivative3 && Sec[i][j]==derivative4)
              {
                case_derivs=4;
              }
              else
              {
                if((Sec[i][j]==derivative1 && Sec[i][j]==derivative2  &&  Sec[i][j]==derivative3)||
                    (Sec[i][j]==derivative1 && Sec[i][j]==derivative2
                     && Sec[i][j]==derivative4) ||
                    (Sec[i][j]==derivative1 && Sec[i][j]==derivative3
                     && Sec[i][j]==derivative4) ||
                    (Sec[i][j]==derivative2 && Sec[i][j]==derivative3
                     && Sec[i][j]==derivative4))

                {
                  case_derivs=3;
                }
                else
                {

                  if( (Sec[i][j]==derivative1 && Sec[i][j]==derivative2) || (Sec[i][j]==derivative1 && Sec[i][j]==derivative3) ||                    (Sec[i][j]==derivative1 && Sec[i][j]==derivative4) || (Sec[i][j]==derivative2 &&  Sec[i][j]==derivative3 ) || (Sec[i][j]==derivative2 && Sec[i][j]==derivative4)
                      || (Sec[i][j]==derivative3 &&  Sec[i][j]==derivative4))
                  {
                    case_derivs=2;
                  }
                  else
                  {
                    if( Sec[i][j]==derivative1 || Sec[i][j]==derivative2 || Sec[i][j]==derivative3 ||
                        Sec[i][j]==derivative4 )
                    {
                      case_derivs=1;
                    }
                    else
                    {
                      case_derivs=0;
                    }
                  }
                }
              }
              if(Sec[i][j]<case_derivs)
              {
                case_derivs=0;
              }
              switch(case_derivs)
              {
              case 0:
                Result4[i][0][derivative1][derivative2][derivative3][derivative4]*=ValuePoints[j][Sec[i][j]];
                break;
              case 1:
                Result4[i][0][derivative1][derivative2][derivative3][derivative4]*=((Complex) (Sec[i][j]))*ValuePoints[j][Sec[i][j]-1];
                break;
              case 2:
                  Result4[i][0][derivative1][derivative2][derivative3][derivative4]*=((Complex) ((Sec[i][j])*(Sec[i][j]-1)))*ValuePoints[j][Sec[i][j]-2];
                break;
              case 3:
                  Result4[i][0][derivative1][derivative2][derivative3][derivative4]*=((Complex) ((Sec[i][j])*(Sec[i][j]-1)*(Sec[i][j]-2)))*ValuePoints[j][Sec[i][j]-3];
                break;
              case 4:
                  Result4[i][0][derivative1][derivative2][derivative3][derivative4]*=((Complex) ((Sec[i][j])*(Sec[i][j]-1)*(Sec[i][j]-2)*(Sec[i][j]-3)))*ValuePoints[j][Sec[i][j]-4];
                break;
              }

            }
          }
        }
      }
    }

  }*/
}



///Could be useful
/*
double Deriv_sAlpha_pointwise(const int size,const int NS,const VVector<int> &Bs, const VVector<int> &Bl, const VVector<int> &BsL,const vvectorI &L,const MMatrix<double>&vi, const MMatrix<Complex> &Result0,const MMatrix<Complex> &Result1,const MMatrix<MMatrix<Complex> > &Result2,MMatrix<Complex> &First_Order,MMatrix<Complex> &Second_Order)
{
   MMatrix<Complex> Deriv_sABpointwise(L.size(),size);
   MMatrix< MMatrix<Complex> > Deriv2_sABpointwise(L.size(),1);
   for(uint k=0;k<L.size();k++)
   {
      Deriv2_sABpointwise[k][0].resize(size,size);
   }
   //   cout << "\n "<< L.size() << " "<< Pairs.size()<< "  "<< NS<< endl;
   //contains the derivative of the Bergman wrt to the variable between 0 and N
   MMatrix<Complex> temp(Bs.size(),size);
   MMatrix<Complex> temp2(L.size(),1);
   MMatrix<Complex> temp3(Bs.size(),1);
   for(uint j=0;j<size;j++)
   {
      for(uint i=0; i<L.size(); i++)
      {
         Deriv_sABpointwise[i][j]= Result1[L[i][0]][j]*conj(Result0[L[i][1]][0]);
         //*conj(Result1[L[i][1]][j]);
 
      }
   }
   for(uint k=0;k<L.size();k++)
   {
      for(uint j=0;j<size;j++)
      {
         for(uint i=0; i<size; i++)
         {
            Deriv2_sABpointwise[k][0][i][j]=Result1[L[k][0]][i]*conj(Result1[L[k][1]][j]);
            //Result2[L[k][0]][0][i][j]*conj(Result2[L[k][1]][0][i][j]);
         }
      }
      //    cout << "\n Derviv2 [k] \n";
      //    show(Deriv2_sABpointwise[k][0]);
   }
 
   temp = average2(Deriv_sABpointwise, Bs, Bl);
   MMatrix<Complex> value_derivatives(size,1);
 
   for(uint l=0;l<size;l++)
   {
      for(int j=0; j<Bs.size(); j++)
      {
         value_derivatives[l][0]+= (Bl[j]<NS) ?  ( (double) BsL[j]) * temp[j][l] * vi[j][0]  : (2* ((double) BsL[j]) * temp[j][l] * vi[j][0]);
 
      }
   }
   MMatrix<Complex> value_derivatives2(size,size);
   for(uint j=0;j<size;j++)
   {
      for(uint i=j; i<size; i++)
      {
         for(uint k=0;k<L.size();k++)
         {
            temp2[k][0]=Deriv2_sABpointwise[k][0][i][j];
         }
         temp3=average2(temp2,Bs,Bl);
         //  cout << "\n Temp3 \n";
         //show(temp3);
         for(uint p=0;p<Bs.size();p++)
         {
            value_derivatives2[i][j]+=(Bl[p]<NS) ?  ( double(BsL[p]) * temp3[p][0] * vi[p][0] ) : (2* (double (BsL[p])) * temp3[p][0] * vi[p][0]);
         }
         if(i>j)
         {
            value_derivatives2[j][i]=conj(value_derivatives2[i][j]);
         }
      }
   }
 
   Complex bergman_sum=0;
   MMatrix<Complex> sAB(L.size(),1);
   for(uint i0=0; i0<L.size(); i0++)
   {
      sAB[i0][0]= Result0[L[i0][0]][0]*conj(Result0[L[i0][1]][0]);
   }
   temp3 = average2(sAB, Bs, Bl);
   for(int j=0; j<Bs.size(); j++)
   {
      bergman_sum+= (Bl[j]<NS) ?  ( double(BsL[j]) * temp3[j][0] * vi[j][0] ) : (2* (double (BsL[j])) * temp3[j][0] * vi[j][0]);
 
   }
   //return(real(value_derivatives));
   First_Order=value_derivatives; //not the most efficient
   Second_Order=value_derivatives2;
   return(real(bergman_sum));
}
 
*/


void ResultsStore::store_Results(const int NS_n,const int K_n,const vector<vector<int> > Sec_n,const VVector<int> &Bs_n,const VVector<int> &Bl_n,const VVector<int> &L_n,const vector<vector<int> > Pairs_n,const MMatrix <Complex> &vi_n)
{
  NS=NS_n;
  K=K_n;
  Sec=Sec_n;
  Bs=Bs_n;
  Bl=Bl_n;
  L=L_n;
  Pairs=Pairs_n;
  vi=vi_n;
}


void Matrices_for_computations::store_Results(const MMatrix<Complex> G_n,const VVector<Complex>
    df_n,const MMatrix<Complex> ddf_n, const MMatrix<Complex>  DxisA_n,const vector<vector<int> > Se_n,const VVector<Complex>
    sAy_n, const VVector<Complex>
    DxP_n,const VVector<Complex>
    cDxP_n,const MMatrix<Complex>  DbD2xP_n,const MMatrix<Complex>  g_n)
{
  G=G_n;
  df=df_n;
  ddf=ddf_n;
  DxisA=DxisA_n;
  Se=Se_n;
  sAy=sAy_n;
  DxP=DxP_n;
  cDxP=cDxP_n;
  DbD2xP=DbD2xP_n;
  g=g_n;

}
void Matrices_for_computations::clear(){}
void ResultsStore::clear()
{
  //TODO
}
