/***************************************************************************
 *   Copyright (C) 2007 by Julien Keller                                   *
 *   jkeller@imperial.ac.uk                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef CURVATURES_H
#define CURVATURES_H

#include "NPFile.h"
#include "Equation.h"
#include "in-output.h"
#include "GenericFunctions.h"
#include "methods.h"
#include "iPoint.h"
#include <complex>
#include "m_matrix2.h"
///#include "Group.h"
#define Complex  std::complex<double>

#define pow2(x) (x*x)
void MMatrixD2fm(Complex y[],const Complex Psi,const int Degree,const int N,MMatrix<Complex> &ddf);

Complex Der2SA(Complex y[], vector<int> sA, int i, int j);
//void FDx_ix_jy_d(const Vector &df, const Matrix &ddf,Matrix &m);
//void FDy_psA(Complex y[], const vector<vector<int> > &sA, int NS);
vector<vector<int> > sAaff( vector<vector<int> > sA, int k);

//void MMatrixInvertF(const MMatrix<Complex> &M, MMatrix<Complex> &Mi, double ERR);

void Computations_of_sections(Complex y[],const int K,MMatrix<Complex> &Result_sections);

//void Fg(const Vector &DxP,const Vector &cDxP,Matrix &DbD2xP,Matrix &g);
//Complex FP(const int NS,const Vector &sAy,const MatrixD &G);
Complex MMatrixFPbis(const int NS,const MMatrix<Complex> &sAy,const MMatrix<Complex> &G);
//Complex DerSA(Complex y[], vector<int> sA, int i);
void MMatrixsA_ybis(const vector<vector<int> > &sA,const int NS,MMatrix<Complex> &sAy,const MMatrix<Complex> &Result_sections);
void VVectorsA_ybis(const vector<vector<int> > &sA,const int NS,VVector<Complex> &sAy,const MMatrix<Complex> &Result_sections);

//void FDy_py_qsA(Complex y[], const vector<vector<int> > &sA,VectorM &w);
//void FDx_ix_jsA(Complex y[], const Vector &df, const Matrix &ddf, const vector<vector<int> > &sA,VectorM &DxixjsA);
//void D2_xP(const Matrix &G, const VectorM &DxixjsA, const Vector &sAy, int NS, Matrix &D2xP);
//void DDbD3_xP(const Matrix &G, const VectorM &DxixjsA, const Matrix &DxisA, int NS,VectorM &DDbD3xP);
void MMatrixDDbD3_xP(const MMatrix<Complex> &G, const MMatrix<MMatrix<Complex> > &DxixjsA, const MMatrix<Complex> &DxisA, int NS,MMatrix<MMatrix<Complex> > &DDbD3xP);


void MMatrixDDbDD4_xP(const MMatrix<Complex> &G, const MMatrix<MMatrix<Complex> > &DxixjsA, int NS,MMatrix<MMatrix<Complex> > &D4xP);


//void D1_xg(const Complex P, const Vector &DxP,const Matrix &D2xP,const Matrix &DbD2xP,const VectorM &DDbD3xP,VectorM &D1xg);
void MMatrixD1_xg(const Complex P,const MMatrix<Complex> &DxP,const MMatrix<Complex> &D2xP,const MMatrix<Complex> &DbD2xP,const MMatrix< MMatrix<Complex> > &DDbD3xP,  MMatrix<Complex> &D1xg);


//void DD2_xg(Complex P, const Vector &DxP,const Matrix &D2xP,const Matrix &DbD2xP,const VectorM &DDbD3xP,const MatrixM &D4xP,MatrixM &DD2xg);
void MMatrixDD2_xg(Complex P, const MMatrix<Complex> &DxP,const MMatrix<Complex> &D2xP,const MMatrix<Complex> &DbD2xP,const MMatrix< MMatrix<Complex> > &DDbD3xP,const MMatrix< MMatrix<Complex> > &D4xP,MMatrix< MMatrix<Complex> > &DD2xg);

void MMatrixFDx_iy_p(const MMatrix<Complex> &df, MMatrix<Complex> &v_deriv);




//void D_xP(const int NS,const Matrix &DxisA,const MatrixD &G,const Vector &sAy,Vector &DxP);

double Find_Ricci_FAST(const double ERR,const Point Pt,const MMatrix<Complex> &G,const int NS,const int K,const vector<vector<int> > &Sec,const Complex Psi,MMatrix<Complex> &g,
                       MMatrix<Complex> &df,MMatrix<Complex> &ddf, int k,Complex y[],MMatrix<Complex> &DxisA,MMatrix<MMatrix<Complex> > &DxixjsA,vector<vector <int> > &Se,MMatrix<Complex> &sAy, MMatrix<Complex> &DxP, MMatrix<Complex> &DbD2xP,MMatrix<Complex> &D2xP,MMatrix<Complex> &Ric,VVector< MMatrix<Complex> > &DDbD3xP,MMatrix < MMatrix<Complex> > &D4xP,MMatrix<Complex> &v_deriv);

void Global_Find_Ricci_FAST(const double ERR,vector<Point> &X,const MMatrix<Complex> &G,const int NS,const int K,const vector<vector<int> > Sec,const Complex Psi,const int Number_Points);



double Deriv_sAlpha_pointwise(const int size,const int NS,const VVector<int> &Bs, const VVector<int> &Bl, const VVector<int> &BsL,const vvectorI &L,const MMatrix<Complex>&vi, const MMatrix<Complex> &Result0,const MMatrix<Complex> &Result1,const MMatrix<MMatrix<Complex> > &Result2,MMatrix<Complex> &First_Order,MMatrix<Complex> &Second_Order);

void Deriv_pointwise(Point &X,const vector<vector<int> > &Sec,const int K,MMatrix<Complex> &Result0,MMatrix<Complex> &Result1,MMatrix< MMatrix<Complex> >&Result2, MMatrix< MMatrix<Complex> >&Result3, MMatrix< MMatrix< MMatrix< Complex> > >&Result4, MMatrix<Complex> &ValuePoints);


class ResultsStore //class to store useful intemediate results from algorithms
{
public:
  ResultsStore() {}
  ResultsStore(const ResultsStore & copied): NS(copied.NS), K(copied.K), Sec(copied.Sec), Bs(copied.Bs), Bl(copied.Bl), L(copied.L),Pairs(copied.Pairs),vi(copied.vi) {}
  ~ResultsStore() {}
  MMatrix<Complex> Results_sections;
  MMatrix<Complex> sA;
  vector<vector<int> > Se;
  VVector<Complex> Temp;
  const int & getNS() const {return NS;}
  const int & getK() const {return K;}
  const vector<vector<int> > & getSec() const {return Sec;}
  const VVector<int> & getBs() const {return Bs;}
  const VVector<int> & getBl() const {return Bl;}
  const VVector<int> & getL() const {return L;}
  const vector<vector<int> > & getPairs() const {return Pairs;}
  const MMatrix <Complex> & getvi() const {return vi;}

  void store_Results(const int NS_n,const int K_n,const vector<vector<int> > Sec_n,const VVector<int> &Bs_n,const VVector<int> &Bl_n,const VVector<int> &L_n,const vector<vector<int> > Pairs_n,const  MMatrix <Complex> &vi_n);
  void clear(); //clears all the content of the store


protected:
  int NS;
  int K;
  vector<vector<int> > Sec;
  VVector<int> Bs;
  VVector<int> Bl;
  VVector<int> L;
  vector<vector<int> > Pairs;
  MMatrix <Complex> vi;
};



//TODO : does not work for affectations ,i.e store functions
class Matrices_for_computations //class to store useful intemediate results from algorithms
{
public:
  Matrices_for_computations() {}
  Matrices_for_computations(const Matrices_for_computations & copied): G(copied.G), df(copied.df), ddf(copied.ddf), DxisA(copied.DxisA), Se(copied.Se), sAy(copied.sAy),
  DxP(copied.DxP),cDxP(copied.cDxP),DbD2xP(copied.DbD2xP),g(copied.g) {}
  ~Matrices_for_computations() {}

  const MMatrix<Complex> & getG() const {return G;}
  const VVector<Complex>
  & getdf() const {return df;}
  const MMatrix<Complex> & getddf() const {return ddf;}
  const MMatrix<Complex> & getDxisA() const {return DxisA;}
  const vector<vector<int> > & getSe() const {return Se;}
  const VVector<Complex>  & getsAy() const {return sAy;}
  const VVector<Complex>  & getDxP() const {return DxP;}
  const VVector<Complex>  & getcDxP() const {return cDxP;}
  const MMatrix<Complex> & getDbD2xP() const {return DbD2xP;}
  const MMatrix<Complex> & getg() const  {return g;}

  void store_Results(const MMatrix<Complex> G_n,const VVector<Complex>
                     df_n,const MMatrix<Complex>ddf_n, const MMatrix<Complex>DxisA_n,const vector<vector<int> > Se_n,const VVector<Complex>
                     sAy_n, const VVector<Complex>
                     DxP_n,const VVector<Complex>
                     cDxP_n,const MMatrix<Complex>DbD2xP_n,const MMatrix<Complex>g_n);
  MMatrix<Complex> & storeG() {return G;}
  void storeG(const MMatrix<Complex> G_n){G=G_n;};
  void storedf(const VVector<Complex> df_n){df=df_n;};
  void storeddf(const MMatrix<Complex>ddf_n){ddf=ddf_n;};
  void storeDxisA(const MMatrix<Complex>DxisA_n){DxisA=DxisA_n;};
  void storeSe(const vector<vector<int> > Se_n){Se=Se_n;};
  void storesAy(const VVector<Complex> sAy_n){sAy=sAy_n;};
  void storeDxP(const VVector<Complex> DxP_n){DxP=DxP_n;};
  void storecDxP(const VVector<Complex> cDxP_n){cDxP=cDxP_n;};
  void storeDbD2xP(const MMatrix<Complex>DbD2xP_n){DbD2xP=DbD2xP_n;};
  void storeg(const MMatrix<Complex>g_n){g=g_n;};
  void clear(); //clears all the content of the store


protected:
  MMatrix<Complex> G; //this contains the hermitian matrice
  VVector<Complex> df;
  MMatrix<Complex>ddf;
  MMatrix<Complex>DxisA;
  vector<vector<int> > Se;
  VVector<Complex> sAy;
  VVector<Complex> DxP;
  VVector<Complex> cDxP;
  MMatrix<Complex> DbD2xP;
  MMatrix<Complex> g;
};

//double VolumeForm_FSmetric2_FAST(const Point &Pt, ResultsStore additResults,MatrixD &G,Vector &df,Matrix &ddf,Matrix &DxisA,vector<vector<int> > &Se,Vector &sAy,Vector &DxP,Matrix &DbD2xP,Matrix &g,Matrix &FS,Matrix&O,const Complex Psi);


double VolumeForm_FSmetric2_FAST2(const Point &Pt, ResultsStore &additResults,MMatrix<Complex> &G,MMatrix<Complex> &df,MMatrix<Complex> &ddf,MMatrix<Complex> &DxisA,vector<vector<int> > &Se,MMatrix<Complex> &sAy,MMatrix<Complex> &DxP,MMatrix<Complex> &DbD2xP,MMatrix<Complex> &g,MMatrix<Complex> &FS,MMatrix<Complex> &O,MMatrix<Complex> &v_deriv,const Complex Psi);

#endif

