/***************************************************************************
 *   Copyright (C) 2007 by Julien Keller                                   *
 *   jkeller@imperial.ac.uk                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Equation.h"



#define pow2(x) (x*x)
#define pow3(x) (x*x*x)
#define pow4(x) (pow2(x*x))
#define pow5(x) (x*pow4(x))
#define pow6(x) (pow2(pow3(x)))
#define pow7(x) (x*pow6(x))
#define pow8(x) (pow2(pow4(x)))




///The Fermat quintic

int degree_polynomial_f(){
  return(N);
}

Complex f(Point y,const Complex Psi)    // the defining polynomial
{

    Complex s=0;
    for(int i=0; i<y.length(); i++) {s += pow(y.comp[i],N);}
    return s-double(N)*Psi*y.comp.product();
}
  
Complex Df(int j, Complex y[],Complex Psi)
{           // Affine derivative of f wrt to x_j  on C^D
    Complex p=1.0;

    for(int i=0; i<Degree; i++) if (i!=j) p*=y[i];
    return double(N)*( pow(y[j],N-1)-Psi*p );
}  


Complex D2f(int j, int k, Complex y[],const Complex Psi,const int Degree,const int N)
{           // Affine derivative of f wrt to x_j & x_k on C^D
    Complex p=1.0;
    if(j==k) p = (Complex(N-1)) * pow(y[j],N-2);
    else {
        for(int i=0; i<Degree; i++) if (i!=j && i!=k) p *= y[i];
        p *= -Psi;    }
        return (Complex(N))*p;
  

}


////////////////////////////////////
///The double cover of an octic////
///////////////////////////////////
/*
int degree_polynomial_f(){
  return(8);
}

  
Complex f(Point y,const Complex Psi)    // the defining polynomial
{

  Complex s=0;
  //s=4.0*pow2(y.comp[0]);
  s=0.5*pow2(y.comp[0]);
  for(int i=1; i<y.length(); i++) {s += pow8(y.comp[i]);}
  //return (s-(8.)*Psi*y.comp.product())/8.;
  return (s-Psi*y.comp.product());
  
}

  
Complex Df(int j, Complex y[],Complex Psi)
{           // Affine derivative of f wrt to x_j  on C^D
  Complex p=1.0;
  for(int i=0; i<Degree; i++){
        if (i!=j) p*=y[i];
  }
  if(j==0){
    return( y[0]-Psi*p);
  }
  if(j!=0){
    return( pow7(y[j])-Psi*p);
  }
  
}

  
Complex D2f(int j, int k, Complex y[],const Complex Psi,const int Degree,const int N)
{           // Affine derivative of f wrt to x_j & x_k on C^D
  Complex p=1.0;
  if(j==k)
  {
    if(j==0)
    {
      p=1.;
    }
    else
    {
      p = 7.*pow6(y[j]);
    }
  }
  else
  {
    for(int i=0; i<Degree; i++)
    {
      if (i!=j && i!=k) p *= y[i];
    }
    p *= -Psi;
  }
  return p;
}

*/
////////////////////////////////////////
////functions for any generic manifold
/////////////////////////////////////////

//should not be modified
Point normalisation_proj_space(Point x){
  Point xnew(x.length());
  xnew.mass=x.mass;
  for(uint i=0;i<x.length();i++){
    xnew.comp[i]=pow(x.comp[i],Projective_space[i]);
  }
  return(xnew);
}
double norm_f(Point y){
  double norm=0;
  for(uint i=0;i<y.length();i++){
    norm+=pow(abs(y.comp[i]), 2*Projective_space[i]);
  }
  return(sqrt(norm));  
}
 
Point scalar_f(Point y,Complex scal){
  Point ynew(y.length());
  ynew.mass=y.mass;
  for(int i=0; i<y.length(); i++){
//    ynew.comp[i]=y.comp[i]*pow(scal,1./((double) Projective_space[i]));
      ynew.comp[i]=y.comp[i]*pow(scal,((double) Projective_space[i]));
  }
  return(ynew);
  
}


Point normalisation_f(Point y) 
{
  
  //  Point ynew(y.length());
  //  ynew.mass=y.mass;
    double norm=norm_f(y);
/*    for(int i=0; i<y.length(); i++){
        ynew.comp[i]=y.comp[i]/pow(norm,1./((double) Projective_space[i]));
    }
    */
    return(scalar_f(y,1./norm));
}

