/***************************************************************************
 *   Copyright (C) 2007 by Julien Keller                                   *
 *   jkeller@imperial.ac.uk                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef EQUATION_H
#define EQUATION_H
#include <complex>
#include "iPoint.h"
#define Complex  std::complex<double>
#include "MMatrix.h"
#include "m_matrix2.h"

Complex f(Point y,const Complex Psi);
int degree_polynomial_f();
Complex D2f(int j, int k, Complex y[],const Complex Psi,const int Degree,const int N);
Complex Df(int j, Complex y[],Complex Psi);
const int Projective_space[5] = { 1 , 1 , 1 , 1 , 1 }; //Quintic fermat
///const int Projective_space[5] = { 4 , 1 , 1 , 1 , 1 }; //Octic



Point normalisation_proj_space(Point x);
Point normalisation_f(Point y);
double norm_f(Point y);
Point scalar_f(Point y,Complex scal);


/// contains all the important data computed during the iterations
//TODO should be somewhere else + comments on all the variables
struct Data_in{
    vector<vector<int> > Pairs;//this is an array of array that contains couples (i,j) that represent the same value in the big hermitian matrice representing the FS metric.  Size around 1000 * 50 *(int,int).
    vector<vector<int> > Sec; // contains the sections s_\alpha=z_\alpha_1....z_\alpha_j
  int BS;
  int NS;
  vector<Point> X;  // Contains the points of the manifolds with their masses
  VVector<int> Bl0; // contains the block up to some simplications
  VVector<int> Bl;
  VVector<int> Bs0;
  VVector<int> Bs;
  vector<vector<int> > PairsL; 
  MMatrix<Complex> vI_new2; //new matrice 
  MMatrix<Complex> sApointwise2; // contains the value of the sections at a points of the manifold
  MMatrix<Complex> v_new2;
  MMatrix<Complex> sABpointwise2; //contains the scalar product of sections at a point
  MMatrix<Complex> ValuePoints;  //contains the value of the powers x^j for a point x.
  MMatrix<Complex> G_new, Gi_new;
  MMatrix<Complex> G; //this contains the hermitian matrice
  MMatrix<Complex> Eta; // contains an error matrice to improve the computation for refined algorithm
  MMatrix<double> Extra_Masses;//this contains the denom at the previous iteration. used for refinment
  MMatrix<Complex> vI_new_old_ratio; //experimental
  MMatrix<Complex>  vI_new_horner; //experimental in order to store smaller numbers
  double ratio_refined; //ratio of points computed for refined version.
  double print_ratio; //VolumeFormFSmetric/Omega. Should be close to 1 for Omega-balanced metric.
};
#endif



