/***************************************************************************
 *   Copyright (C) 2007 by Julien Keller   *
 *   j.keller@imperial.ac.uk   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// This file contains some basic operations for the whole program

#include "GenericFunctions.h"
//#include "NPFILE.hpp"
/*
#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>

#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/lu.hpp>
*/
#define Complex  std::complex<double>
#define pow2(x) (x*x)
#define pow3(x) (x*x*x)
#define pow4(x) pow2(x*x)
#define pow5(x) (x*pow4(x))
#define pow6(x) pow2(pow3(x))
#define pow8(x) pow2(pow4(x))


int ChooseMax(const Point &x) // finds the largest abs value coordinate
{                       //  and go to that affine patch on P^D
  int M=0;
  double big=abs(x.comp[0]);
  for (int i=1; i<N; i++)
  {
    if(abs(x.comp[i])>big)
    {
      M=i;
      big=abs(x.comp[i]);
    }
  }
  return M;
}




void AffineCoord(const Point &x, Complex y[], int M)
{     // gives the Affine coordinates instead
  ///Point xnew=scalar_f(x,1./pow(x.comp[M],1./Projective_space[M]));
  Point xnew=scalar_f(x,1./x.comp[M]);
  
  int i;
  for(i=0; i<M; i++)   y[i]   = xnew.comp[i];
  for(i=M+1; i<N; i++) y[i-1] = xnew.comp[i];
}


VVector<Complex> ChooseY4(Complex y[],const Complex Psi) // Choose y4, i.e. the coord on P^N that is not used on X, and reorder
{                 // so that y4 is the last one. This is another point on the S_N orbit!
  // Also: returns new d_i f at y, and reorders the array y (checked)
  Complex T;
  VVector<Complex> fi;
  fi.resize(Degree,1);
  fi[0]=Df(0,y,Psi);
  double big=abs(fi[0]);
  int M=0;
  for (int i=1; i<Degree; i++) {fi[i]=Df(i,y,Psi); if(abs(fi[i])>big) {M=i; big=abs(fi[i]);}}
  //switching
  T=y[M]; y[M]=y[Dimension]; y[Dimension]=T;
  T=fi[M]; fi[M]=fi[Dimension]; fi[Dimension]=T;
  return fi;
}

void ChooseY4Fast(Complex y[],const Complex Psi) // Choose y4, i.e. the coord on P^N that is not used on X, and reorder
{                 // so that y4 is the last one. This is another point on the S_N orbit!
  // Also: returns new d_i f at y, and reorders the array y (checked)
  Complex T;
  VVector<Complex> fi;
  double big=abs(Df(0,y,Psi));
  int M=0;
  Complex res;
  for (int i=1; i<Degree; i++) {
      res=Df(i,y,Psi); 
      if(abs(res)>big) 
      {
         M=i; big=abs(res);
      } 
  }
  //switching
  T=y[M]; y[M]=y[Dimension]; y[Dimension]=T;
//  T=fi[M]; fi[M]=fi[Dimension]; fi[Dimension]=T;
//  return fi;
}



VVector<Complex> ProjChooseY4(Complex y[],const Complex Psi,VVector<int> &VProjective_space) // Choose y4, i.e. the coord on P^N that is not used on X, and reorder
{                 // so that y4 is the last one. This is another point on the S_N orbit!
  // Also: returns new d_i f at y, and reorders the array y (checked)
  Complex T;
  VVector<Complex> fi;
  fi.resize(Degree,1);
  fi[0]=Df(0,y,Psi);
  double big=abs(fi[0]);
  ///pow(abs(fi[0]),VProjective_space[0]);
  int dim=Degree-1;
  int M=0;
  for (int i=1; i<Degree; i++)
  {
    fi[i]=Df(i,y,Psi);
    if(abs(fi[i])>big)
      ///      (pow(abs(fi[i]),VProjective_space[i])>big)
    {
      M=i;
      big=(abs(fi[i]));
      ///big=pow(abs(fi[i]),VProjective_space[i]);
    }
  }
  //switching
  if(M!=dim)
  {

    T=y[M];
    y[M]=y[dim];
    y[dim]=T;
    ///    y[M]=pow(y[dim],((double) VProjective_space[dim])/((double) VProjective_space[M]));
    ///    y[dim]=pow(T,((double) VProjective_space[M])/((double) VProjective_space[dim]));

    T=fi[M];
    /// fi[M]=pow(fi[dim],((double) VProjective_space[dim])/((double) VProjective_space[M]));
    fi[M]=fi[dim];
    /// fi[dim]=pow(T,((double) VProjective_space[M])/((double) VProjective_space[dim]));
    fi[dim]=T;

    int temp=VProjective_space[M];
    VProjective_space[M]=VProjective_space[dim];
    VProjective_space[dim]=temp;
  }
  return fi;
}


MMatrix<Complex> MMatrixChooseY4(Complex y[],const Complex Psi) // Choose y4, i.e. the coord on P^N that is not used on X, and reorder
{                 // so that y4 is the last one. This is another point on the S_N orbit!
  // Also: returns new d_i f at y, and reorders the array y (checked)
  Complex T;
  MMatrix<Complex> fi(Degree,1);
  fi[0][0]=Df(0,y,Psi);
  double big=abs(fi[0][0]);
  int M=0;
  for (int i=1; i<Degree; i++) {fi[i][0]=Df(i,y,Psi); if(abs(fi[i][0])>big) {M=i; big=abs(fi[i][0]);}}
  //switching
  T=y[M]; y[M]=y[Dimension]; y[Dimension]=T;
  T=fi[M][0]; fi[M][0]=fi[Dimension][0]; fi[Dimension][0]=T;
  return fi;
}

//transforms a vectorD to a MMatrix
//Other programs to the same stuff
/*VectorD VectorDtoMMatrix(const MMatrix<double> vI_new2)
{
  VectorD v(vI_new2.getHeight());
  for(uint i=0;i<vI_new2.getHeight();i++)
  {
    v(i)=vI_new2[i][0];
  }
  return(v);
}
*/

void PutToFullMatrix(const MMatrix<Complex> v,MMatrix<Complex> &M, const vvectorI &L, const VVector<int> &Bs, const VVector<int> &Bl)
{
  int j;
  // implicitly assumes that the incoming matrix has ONLY entries that are on the list
  for(uint i=0; i<Bl.size(); i++)
  {
    for(j=Bl[i]; j<Bl[i]+Bs[i]; j++)
    {
      M[L[j][0]][L[j][1]] = v[i][0];
      if(L[j][0]!=L[j][1]) M[L[j][1]][L[j][0]] = v[i][0]; ///TODO is it useful ??
    }
  }
}
/*
void PutToFullVector(const MMatrix<Complex> v,VVector<Complex> &v_out, const vvectorI &L, const VVector<int> &Bs, const VVector<int> &Bl)
{
  int j;
  // implicitly assumes that the incoming matrix has ONLY entries that are on the list and v_out has the right size
  for(uint i=0; i<Bl.size(); i++)
  {
    for(j=Bl[i]; j<Bl[i]+Bs[i]; j++)
    {
      //M[L[j][0]][L[j][1]] = v[i][0];
     v_out[j]=v[L[j][0]][0]; 
    }
  }
}
*/


/*
void MVectorToMMatrix2(const MMatrix<Complex> &vI_new2, MMatrix<Complex> &M, const vvectorI &L, VVector<int> &Bs, VVector<int> &Bl)
{
  int w[2], j;
  int size_M=M.getHeight();
  for(int i=0; i<Bl.getHeight(); i++)
  {
    for(j=Bl[i]; j<Bl[i]+Bs[i]; j++)
    {
      w[0]= L[j][0]; w[1]= L[j][1];
      if(w[0]<size_M && w[1]<size_M && i<vI_new2.getHeight())
      { 
        M[w[0]][w[1]] = vI_new2[i][0];
        if(w[0]!=w[1]) M[w[1]][w[0]] = vI_new2[i][0];
      }
    }
    }
}
*/
/*
void vectorToMatrix2(const VectorD &v, MatrixD &M, const vvectorI &L, const VectorI &Bs, const VectorI &Bl, int size_M)
{
  int w[2], j;
 
  for(uint i=0; i<Bl.size(); i++)
  {
    for(j=Bl(i); j<Bl(i)+Bs(i); j++)
    {
      w[0]= L[j][0]; w[1]= L[j][1];
      if(w[0]<size_M && w[1]<size_M && i<v.size())
      { 
        M(w[0],w[1]) = v(i);
        if(w[0]!=w[1]) M(w[1],w[0]) = v(i);
      }
    }
    }
}
*/
/*
void vectorToMatrix(const VVector<double> &v, MMatrix<double> &M, const vvectorI &L, const VectorI &Bs, const VectorI &Bl)
{
  int w[2], j;    // implicitly assumes that the incoming matrix has ONLY entries that are on the list
  for(uint i=0; i<Bl.size(); i++)
    for(j=Bl(i); j<Bl(i)+Bs(i); j++)
    {
      w[0]= L[j][0]; w[1]= L[j][1];
      M[w[0]][w[1]] = v[i];
      if(w[0]!=w[1]) M[w[1]][w[0]] = v[i];
    }
}
*/

/*void MMatrixToMatrix(const MMatrix<Complex> v, MMatrix<Complex> &M, const vvectorI &L, const VVector<int> &Bs, const VVector<int> &Bl)
{
  int j;
  // implicitly assumes that the incoming matrix has ONLY entries that are on the list
  for(uint i=0; i<Bl.size(); i++)
    for(j=Bl[i]; j<Bl[i]+Bs[i]; j++)
    {
      M[L[j][0]][L[j][1]] = v[i][0];
      if(L[j][0]!=L[j][1]) M[L[j][1]][L[j][0]] = v[i][0]; ///TODO is it useful ??
    }
}
*/
/*
void MMatrixToMMatrix(const MMatrix<double> v, MMatrix<double> &M, const vvectorI &L, const VectorI &Bs, const VectorI &Bl)
{
   int j;
   // implicitly assumes that the incoming matrix has ONLY entries that are on the list
   for(uint i=0; i<Bl.size(); i++)
      for(j=Bl(i); j<Bl(i)+Bs(i); j++)
      {
         M[L[j][0]][L[j][1]] = v[i][0];
         if(L[j][0]!=L[j][1]) M[L[j][1]][L[j][0]] = v[i][0]; ///TODO is it useful ??
      }
}
*/

/*
void MatrixToVector(VVector<double> &v, const MMatrix<double> &M, const vvectorI &L, const VectorI &Bs, const VectorI &Bl)
{
  VVector<double> t;
  t.resize(L.size(),1);
  for(int i=0; i<L.size(); i++) t[i] = M[L[i][0]][L[i][1]];
  v = average(t, Bs, Bl);
}
*/

//TODO to remove 
void GetFromMMatrix(MMatrix<Complex> &v,  MMatrix<Complex> &M, const vvectorI &L, VVector<int> &Bs,  VVector<int> &Bl)
{
  MMatrix<Complex> t(L.size(),1);
  for(int i=0; i<L.size(); i++) t[i][0] = M[L[i][0]][L[i][1]];
  v = average2(t, Bs, Bl);
}

void PutToVector(MMatrix<Complex> &v, const MMatrix<Complex> &M, const vvectorI &L, const VVector<int> &Bs, const VVector<int> &Bl)
{
  MMatrix<Complex> t(L.size(),1);
  for(uint i=0; i<L.size(); i++) t[i][0] = M[L[i][0]][L[i][1]];
  v = average2(t, Bs, Bl);
}


//same as before but for MMatrix package.
MMatrix<Complex> initialize2(const vector<vector<int> > &S, VVector<int> &Bl)
{
  vector<int> t;
  MMatrix<Complex> v(Bl.getHeight(),1);
  for(uint i=0; i<Bl.getHeight(); i++) {t=S[Bl[i]]; v[i][0] = (t[0]==t[1]) ? 3.14157*(rand()%10+1.414) : 67.45*(rand()%10+3.14157);}
  //  for(uint i=0; i<Bl.size(); i++) {t=S[Bl(i)]; v[i][0] = (t[0]==t[1]) ? 1.0 : 0.1*(rand()%10+1);}

  //making the assumptions that diagonal terms are the most important ones. Not clearly optimal...
  return v;
}

/*
Complex determ(const MMatrix<Complex> &O)
{
  Complex z;
  if(Dimension==1) z = O(0,0);
  if(Dimension==2) z = O(0,0)*O(1,1) - O(0,1)*O(1,0);
  if(Dimension==3) z = O(1,1)*O(2,2)*O(0,0)-O(1,1)*O(2,0)*O(0,2)+O(2,1)*O(0,2)*O(1,0)
                         -O(2,1)*O(1,2)*O(0,0)+O(0,1)*O(1,2)*O(2,0)-O(0,1)*O(2,2)*O(1,0);
  return z;
}
*/
void MMatrixFSmetric(Complex y[],MMatrix<Complex> &FS)
{            // Fubini-Study metric components
  int i,j;
  double denom=1.0;
  //  u::identity_matrix<Complex> Id(Degree);
  for (j=0; j<Degree; j++) denom += norm(y[j]);
  FS.fillWith(0);
  for (i=0; i<Degree; i++)
  {
    FS[i][i]=1.0;
    for (j=i; j<Degree; j++)
    {

      FS[i][j] -= conj(y[i])*y[j]/denom; // important to conj!
      if(i!=j) FS[j][i] = conj(FS[i][j]);
    }
  }
  FS*=(1.0 / denom);
}

void MMatrixProjFSmetric(Complex y[],MMatrix<Complex> &FS,VVector<int> &VProjective_space)
{            // Fubini-Study metric components
  int i,j;
  double denom=1.0;
  //  u::identity_matrix<Complex> Id(Degree);
  for (j=0; j<Degree; j++) denom += (abs(pow(y[j],2*VProjective_space[j])));
  FS.fillWith(0);
  for (i=0; i<Degree; i++)
  {
    FS[i][i]=pow2(VProjective_space[i])*pow(abs(y[i]),2*(VProjective_space[i]-1));
    for (j=i; j<Degree; j++)
    {

      FS[i][j] -= ((double) VProjective_space[i]*VProjective_space[j])*conj(pow(y[i],VProjective_space[i]))*pow(y[j],VProjective_space[j])*pow(y[i],VProjective_space[i]-1)*pow(conj(y[j]),VProjective_space[j]-1)/denom;
      //one should do the case where the coeff is zero to simplify
      // important to conj!
      if(i!=j) FS[j][i] = conj(FS[i][j]);
    }
  }
  FS*=(1.0 / denom);
}


/*
Complex getOneVol(const Point &x, const Matrix &g,const Complex Psi)
{
  Complex y[Degree];
  int i=ChooseMax(x), j;
  AffineCoord(x, y, i);
  Complex numer, denom;
  Matrix FS(Degree,Degree), O(Dimension,Dimension);
  Vector df(Degree), d4(Dimension);
  df = ChooseY4(y,Psi);
  FS = FSmetric(y);
  for (i=0; i<Dimension; i++) d4(i) = -df(i)/df(Dimension);
  for (i=0; i<Dimension; i++)
    for (j=i; j<Dimension; j++)
    {
      O(i,j) = FS(i,j) + FS(Dimension,j)*d4(i) + (FS(i,Dimension) + FS(Dimension,Dimension)*d4(i)) * conj(d4(j));
      if(i!=j) O(j,i) = conj(O(i,j));
    }
 
  denom = determ(O);
  //   cout << "Matrice G \n "<< g;
  numer = determ(g);
  return numer / denom;
}
*/




extern "C"
{
  //void zgetrf_(const int &N, const int &N, cx *, const int &N, int * ,       const int &N);
  //void zgetri_(const int &N,               cx *, const int &N, int * , cx *, const int &N, const int &N);
  //void zhetrf_(const char &UPLO, const int &N, Complex *, const int &N, int *, Complex *, const int &LWORK, const int &INFO);
  //void zhetri_(const char &UPLO, const int &N, Complex *, const int &N, int *, Complex *, const int &INFO);
  void zhetrf_(const char &, const int &, Complex *, const int &, int *, Complex *, const int &, const int &);
  void zhetri_(const char &, const int &, Complex *, const int &, int *, Complex *, const int &);
}

void Invert2        (const MMatrix<Complex> &M, MMatrix<Complex> &Mi, double ERR)
{
  int i,j, N=M.getHeight();
  int INFO, *IPIV = new int [N];
  Complex  *WORK = new Complex [N], *A = new Complex [N*N], t;
  for(i=0; i<N; i++)
    for(j=i; j<N; j++)
      A[i*N+j] = M[i][j];

  zhetrf_('L',N, A, N, IPIV, WORK, N, INFO);
  zhetri_('L',N, A, N, IPIV, WORK,    INFO);

  for(i=0; i<N; i++)
  {
    Mi[i][i] = (A[i*N+i]);
    for(j=i+1; j<N; j++)
    {
      Mi[i][j] = (A[i*N+j]);
      Mi[j][i] = conj(Mi[i][j]);
    }
  }
  delete [] A;
  delete [] IPIV;
  delete [] WORK;
}



