/***************************************************************************
 *   Copyright (C) 2007 by Julien Keller   *
 *   j.keller@imperial.ac.uk   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef GENERICFUNCTIONS
#define GENERICFUNCTIONS

#include "in-output.h"
#include "iPoint.h"
#include "MMatrix.h"
#include "Equation.h"
#define Complex  std::complex<double>


template<class T> static T averageOne(const VVector<T> &v, const int &L, const int &j)
{                       // L is length of the block, j is starting point

  T s;
  for(int k=j;k<j+L+1;k++)
  {
     s+=v[k];
  } 
//  u::range r(j, j+L);         // start <= index < stop
//  T s = sum(project(v, r));
  return s / (double(L));
}/*

template<class T> VVector<T> average(const VVector<T> &v, const VectorI &Bs, const VectorI &Bl)
{                       // average out a vector
  int BS=Bs.size();
  VVector<T> f;
  f.resize(BS,1);
  for(int j=0; j<BS; j++) f[j] = averageOne<T>(v, Bs(j), Bl(j));
  return f;
}
*/
template<class T> static T averageOne2(const MMatrix<T> &v, const int &L, const int &j)
{                       // L is length of the block, j is starting point
    T s=  v.subMMatrix(j,0,L,1).sum();
    return s / double(L);
}

template<class T> MMatrix<T> average2(const MMatrix<T> &v, const VVector<int> &Bs, const VVector<int> &Bl)
{                       // average out a vector or a Matrice (i.e seen as a collection of vectors)
    MMatrix<T> f(Bs.size(),v.getWidth());
    for(int j=0; j<Bs.size(); j++){
        for(uint k=0;k<v.getWidth();k++){
            f[j][k] = v.subMMatrix(Bl[j],k,Bs[j],1).sum()/((double) Bs[j]);
            }
        }
    return f;
}

template<class T> VVector<T> Vaverage2(const VVector<T> &v, const VVector<int> &Bs, const VVector<int> &Bl)
{                       // average out a vector or a Matrice (i.e seen as a collection of vectors)
    VVector<T> f;
    f.resize(Bs.size(),1);
    for(int j=0; j<Bs.size(); j++){
        {
            f[j] = v.subMMatrix(Bl[j],0,Bs[j],1).sum()/((double) Bs[j]);
            }
        }
    return f;
}

/*
  int BS=Bs.size();
  MMatrix<T> f(BS,1);
  for(int j=0; j<BS; j++) f[j][0] = averageOne2<T>(v, Bs(j), Bl(j));
  return f;
*/
template<class T> int find(const T &a, const vector<T> &V) // find 'a' in a vector V
{
  for(int i=0; i<V.size(); i++) if(a==V[i]) return i;
  return -1;
}  // return -1 if fails (hope to produce an error)


template<class T> T prod(const vector<T> &v) {T p=1; for(int i=0; i<v.size(); i++) p*=v[i]; return p;}

//Complex determ(const MMatrix<Complex> &O);
//TODO Give the details !
MMatrix<Complex> initialize2(const vector<vector<int> > &S, VVector<int> &Bl);
//void vectorToMatrix(const VVector<double> &v, MMatrix<double> &M, const vvectorI &L, const VectorI &Bs, const VectorI &Bl);
//void MatrixToVector(VVector<double> &v, const MMatrix<double> &M, const vvectorI &L, const VectorI &Bs, const VectorI &Bl);

//void vectorToMatrix2(const VectorD &v, MatrixD &M, const vvectorI &L, const VectorI &Bs, const VectorI &Bl, int size_M);
//VectorD VectorDtoMMatrix(const MMatrix<double> vI_new2);
//void MVectorToMMatrix2(const MMatrix<Complex> &vI_new2, MMatrix<Complex> &M, const vvectorI &L, VVector<int> &Bs, VVector<int> &Bl);


/*void MMatrixToMMatrix(const MMatrix<double> v, MMatrix<double> &M, const vvectorI &L, const VectorI &Bs, const VectorI &Bl);*/

void PutToFullMatrix(const MMatrix<Complex> v,MMatrix<Complex> &M, const vvectorI &L, const VVector<int> &Bs, const VVector<int> &Bl);
void PutToVector(MMatrix<Complex> &v, const MMatrix<Complex> &M, const vvectorI &L, const VVector<int> &Bs, const VVector<int> &Bl);
void GetFromMMatrix(MMatrix<Complex> &v,  MMatrix<Complex> &M, const vvectorI &L, VVector<int> &Bs,  VVector<int> &Bl);
/*void PutToFullVector(const MMatrix<Complex> v,VVector<Complex> &v_out, const vvectorI &L, const VVector<int> &Bs, const VVector<int> &Bl);
*/

int ChooseMax(const Point &x);
void AffineCoord(const Point &x, Complex y[], int M);
//Vector ChooseY4(Complex y[],const Complex Psi);
MMatrix<Complex> MMatrixChooseY4(Complex y[],const Complex Psi);
VVector<Complex> ProjChooseY4(Complex y[],const Complex Psi,VVector<int> &VProjective_space);
VVector<Complex> ChooseY4(Complex y[],const Complex Psi);
void ChooseY4Fast(Complex y[],const Complex Psi);
Complex Df(int j, Complex y[],Complex Psi);
void MMatrixProjFSmetric(Complex y[],MMatrix<Complex> &FS,VVector<int> &VProjective_space);
// FSmetric(Complex y[]);
void MMatrixFSmetric(Complex y[],MMatrix<Complex> &FS);
//Complex getOneVol(const Point &x, const Matrix &g,const Complex Psi);

void Invert2(const MMatrix<Complex> &M, MMatrix<Complex> &Mi, double ERR);              
//inverse of a hermitian matrix
#endif


