/***************************************************************************
 *   Copyright (C) 2007 by Julien Keller   *
 *   j.keller@imperial.ac.uk   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef GROUP
#define GROUP

#include <algorithm>
#include <set>

#include <vector>
#include "GenericFunctions.h"


template<class T> vector<T> S_N(int n)
{           // makes the group S_n, T = vector<int>
  vector<T> f;
  T v;
  for(int i=0; i<n; i++) v.push_back(i);
  do f.push_back(v); while(next_permutation(v.begin(), v.end()));
  return f;
}

template<class T> T permute(const T &Perm, const T &v)
{           // applies a permuation Perm to a vector v
  T f;            // T = vector<int>
  for(int i=0; i<v.size(); i++) f.push_back( v[Perm[i]] );
  return f;
}


template<class T> vector<T> SetToVector(const set<T> &s)
{           // convert a set to a vector
  typedef typename set<T>::iterator sIter;
  vector<T> f;
  for(sIter i=s.begin(); i!=s.end(); i++) f.push_back(*i);
  return f;
}

template<class T> set<T> generateOrbit(const T&v, const vector<T> &S_n)
{           // generate S_n orbit of v=(i,j)
  set<T> f; // T=vector<int>
  for(int i=0; i<S_n.size(); i++) f.insert(permute(S_n[i],v) );
  return f;
}


vector<int>  swap(vector<int> t);
vector<int>  order(vector<int> t);

#endif

