


#include "WP_procedures.h"
#define Complex  std::complex<double>
#define pow2(x) (x*x)
#define pow3(x) (x*x*x)
#define pow4(x) pow2(x*x)  



double WPmetric(MMatrix<Complex> Gnew,Data_in & Data,const int Number_Points,Complex Psi,const int K,ResultsStore &additResults,double put_to_memory)
//returns the value of the WP metric at psi/psi
{
  int NS=Data.NS;

  MMatrix<Complex> L(NS,NS), Linv(NS,NS);
  L = Gnew;
  CsqrtF(L);
  Invert2(L,Linv,1e-5); ///what is the required precision for the inverse ?
  MMatrix<Complex> projA(NS,NS);
  projA = projAF(Data.X,Linv,Gnew,Number_Points,K,Data,additResults,Psi,put_to_memory);
  double gpsipsi;
  gpsipsi = gpsipsiF(Data.X,Linv,Gnew,projA,Data,additResults, NS,Number_Points, K, Psi,put_to_memory);
  return(gpsipsi);
}


Complex DerivY4(const Point &X,const Complex Psi)
//return \partial w4/\partial \psi in order to compute later \delta s
{
  Complex prod=1;
  for(int i=0; i<Degree-1;i++)
  {
    prod=prod*X.affine[i];
  }
  return(prod*X.affine[Degree-1]/( (pow4(X.affine[Degree-1]))-Psi*prod));   // because we are considering a degree 5 eqn

  //gives \partial w4/\partial \psi=  w1w2w3w4/(5w4 ^4 - Psi*w1*w2*w3
  //I could also use the function Df if I want more general.
}


void DerivSectionsAffine(const Point &X, const Complex Psi, ResultsStore &additResults,MMatrix<Complex> &sA,double put_to_memory)
// we want to derive all the sections with respect to w4 = max coord in Affine space
//this procedure does not use the balanced metric : if I want to get the derivative wrt to the
// base of balanced sections, one just needs to multiply the matrix obtained by G, the matrix
//constructed using all the Data computed :  PutToFullMatrix(additResults.getvi(),G, additResults.getPairs(), additResults.getBs(), additResults.getBl());
{
  int NS=additResults.getNS();
  //  int K=additResults.getK();
  // Complex y[Degree];
  //  int r = ChooseMax(X);
  //  AffineCoord(X, y, r);
  //  ChooseY4(y,Psi);
  if(put_to_memory==0)
  {
    int r = ChooseMax(X);

    additResults.Se = sAaff(additResults.getSec(), r); //gives the powers (p1,p2,p3,p4) to form the polynomials  w1^p_1 w_2^p_2 w_3^p_3 w_4^p_4  after deriving by w4
    MMatrixsA_ybis(additResults.Se,NS,sA,X.Result_sections); //gives the derivative of the sections wrt w4 at y
  }
  ///could be also contained in X but takes a lot of mem
  //  MMatrix<Complex> Result_sections(Degree,K+1);
  // Computations_of_sections(y,K,Result_sections); //gives the powers at the point y in order to construct later the polynomials expression representing the \delta s, i.e computes generic w1^p_1 w_2^p_2 w_3^p_3 w_4^p_4 with (w1,w2,w3,w4)=y
  //MMatrix<Complex> sAy(NS,1);
  if(put_to_memory==1)
  {
    MMatrixsA_ybis(X.Se,NS,sA,X.Result_sections); //gives the derivative of the sections wrt w4 at y
    //sA=sAy;
  }
  if(put_to_memory==2)
  {
    int r = ChooseMax(X);
    additResults.Se = sAaff(additResults.getSec(), r); //gives the powers (p1,p2,p3,p4) to form the polynomials  w1^p_1 w_2^p_2 w_3^p_3 w_4^p_4  after deriving by w4
    MMatrixsA_ybis(additResults.Se,NS,sA,additResults.Results_sections); //gives the derivative of the sections wrt
   //show(sA);
  // getchar();
  }
}


/// To modify
void CsqrtF( MMatrix<Complex> &Csqrt ) /// Root Square of square matrix H = Csqrt (initial) is Csqrt (final) : H = prod(Csqrt,herm(Csqrt)) ...
{
  Complex sumsq;
  int r = Csqrt.getHeight();
  MMatrix<Complex> H(r, r);

  H.fillWith(Csqrt);
  int c;
  Csqrt.fillWith(0);
 int a,b;
  for( a=0; a<r; a++)
  {
    for( b=0; b<a; b++)
    {
      sumsq=0;
      if(b>0){for(c=0; c <= b-1; c++) sumsq += Csqrt[a][c]*conj(Csqrt[b][c]);};
      Csqrt[a][b]=(H[a][b]-sumsq)/(Csqrt[b][b]);
    }
    sumsq=0;
    if(a>0){for(c=0; c <= a-1; c++) sumsq += Csqrt[a][c]*conj(Csqrt[a][c]);}
    Csqrt[a][a] = sqrt(H[a][a]-sumsq);
  }
  /*show(Csqrt);
  MMatrix<Complex> Save(r,r);
  Save.fillWith(Csqrt);
  Csqrt.transpose();
  MMatrix<Complex> Temp=H-multiply(Save,conj(Csqrt));
  cout<< " Temp = "<< Temp.norm_euclid();
  getchar();*/
}


// Ortho-Normalized S
void LinvS(Point &X, MMatrix<Complex> &Linv,Data_in & Data,ResultsStore &additResults,const int NS,const int K,VVector<Complex> &S)
{
  //S2 is assumed to have the right size
  // S.resize(NS,1);

  sAlpha_pointwise2a ( K,Data,X );

  //VVector<Complex> Temp;
  additResults.Temp.fillWith(0);
  int i,j;
  for( i=0; i<NS; i++)
  {
    for( j=0; j<NS; j++)
    {
      additResults.Temp[i]+=Linv[i][j]*Data.sApointwise2[j][0];
    }
  }
  S.fillWith(additResults.Temp);
//  show(S); getchar();
  /*
  for(int i=0; i<NS; i++)
  {
    S[i]=Data.sApointwise2[i][0];
  }
  S = multiply(Linv, S);*/
//  S=Vaverage2(S,Data.Bs0,Data.Bl0);

}

// Ortho-Normalized deltaS

void LinvdeltaS(const Point &X, MMatrix<Complex> &Linv, VVector<Complex> &deltaS, Data_in & Data, ResultsStore &additResults,const Complex Psi,const int NS,double put_to_memory)
{

  if(put_to_memory==1 || put_to_memory==0)
  {
    deltaS.fillWith(0);
  }
  //MMatrix<Complex> sA(NS,1);
  //VectorDerivSectionsAffine(X,Psi,additResults,deltaS);
  DerivSectionsAffine(X,Psi,additResults,additResults.sA,put_to_memory);
//  show(additResults.sA);

  int i,j;
  Complex factor = DerivY4(X, Psi);
  //additResults.sA*= DerivY4(X, Psi);
// show(additResults.sA);
//show(Linv);
//show(multiply(Linv,additResults.sA));
  for( i=0; i<NS; i++)
  {
    for( j=0; j<NS; j++)
    {
       deltaS[i]+=Linv[i][j]*additResults.sA[j][0];
    }
    deltaS[i]=factor*deltaS[i];
  }
 // deltaS=Vaverage2(deltaS,Data.Bs0,Data.Bl0);

// cout <<" factor" << factor;
 //  show(deltaS);
 //  getchar();

  /*
    for(int i=0; i<NS; i++)
    {
      deltaS[i]=sA[i][0];
    }
    deltaS = multiply(Linv, deltaS); //I did not multiply here by dw4/dpsi;
  */
//   deltaS *=factor;
}


MMatrix<Complex> projAF(vector<Point> &X, MMatrix<Complex> &Linv, MMatrix<Complex> &Ginv, const int Number_Points,const int K,Data_in & Data,ResultsStore &additResults,const Complex Psi,double put_to_memory)
{


  int r,NS=Ginv.getHeight();
  MMatrix<Complex> projA(NS,NS);
  VVector<Complex> SX;
  VVector<Complex> deltaSX;
  MMatrix<Complex> Result_sections(Degree,K+1);
  SX.resize(NS,1);
  deltaSX.resize(NS,1);
  additResults.sA.resize(NS,1);
  additResults.Temp.resize(NS,1);
//  Complex affine_point[Degree];
  Complex D_function;
  for(int n_points=0; n_points<Number_Points; n_points++)
  {
    if ( n_points % ( ( int ) Number_Points/100 ) ==0 )
    {
      cout.flush();
      cout << "*** Computing the projection  "<< ( 100.*n_points ) / ( Number_Points ) << "%  \r" ;
    }
    r = ChooseMax(X[n_points]);
    if(put_to_memory==1)
    {
      AffineCoord(X[n_points],X[n_points].affine, r);
      ChooseY4Fast(X[n_points].affine,Psi);
      X[n_points].Se = sAaff(additResults.getSec(), r);
      X[n_points].Result_sections.resize(Degree,K+1);
      Computations_of_sections(X[n_points].affine,K,X[n_points].Result_sections);
      LinvS(X[n_points],Linv,Data,additResults,NS,K,SX);
      LinvdeltaS(X[n_points],Linv,deltaSX, Data,additResults, Psi,NS,put_to_memory);

    } //gives the powers (p1,p2,p3,p4) to form the polynomials  w1^p_1 w_2^p_2 w_3^p_3 w_4^p_4  after deriving by w4
    if(put_to_memory==0)
    {
      AffineCoord(X[n_points],X[n_points].affine, r);
      ChooseY4Fast(X[n_points].affine,Psi);
      X[n_points].Result_sections.resize(Degree,K+1);
      Computations_of_sections(X[n_points].affine,K,X[n_points].Result_sections); //gives the powers at the point y in order to
      LinvS(X[n_points],Linv,Data,additResults,NS,K,SX);
      LinvdeltaS(X[n_points],Linv,deltaSX,Data,additResults, Psi,NS,put_to_memory);
      //show(deltaSX);getchar();
    }
    if(put_to_memory==2)
    {
      AffineCoord(X[n_points],X[n_points].affine, r);
      ChooseY4Fast(X[n_points].affine,Psi);
      Computations_of_sections(X[n_points].affine,K,Result_sections);
      X[n_points].SX.resize(NS,1);
      X[n_points].deltaSX.resize(NS,1);
      LinvS(X[n_points],Linv,Data,additResults,NS,K,X[n_points].SX);
      additResults.Results_sections=Result_sections;
      LinvdeltaS(X[n_points],Linv,X[n_points].deltaSX,Data,additResults, Psi,NS,put_to_memory);
    }
    X[n_points].D_function = DsAlphaBeta_pointwise2a ( Data ); ///Ginv should be already loaded in Data
    ///note that the point is already loaded thanks to LinvS
    int i,j;
    if(put_to_memory==2)
    {
      for( i=0; i<NS; i++)
      {
        for( j=0; j<NS; j++)
        {
          projA[i][j] += X[n_points].deltaSX[i]*conj(X[n_points].SX[j])*X[n_points].mass/X[n_points].D_function;
        }
      }
    }
    if(put_to_memory==0  || put_to_memory==1)
    {
      for( i=0; i<NS; i++)
      {
        for( j=0; j<NS; j++)
        {
          //cout << "\n infos : "<< deltaSX[i]<< "  "<< SX[j]<< "  " << X[n_points].D_function;
          projA[i][j] += deltaSX[i]*conj(SX[j])*X[n_points].mass/X[n_points].D_function;
        }
      }
    }
  }

  projA*=1./(Number_Points);  ///*=N/Volume  //TODO different from Segio's code
//  show(projA);getchar();
   return projA;

}





void PdeltaLinvS(Point &X, MMatrix<Complex> &Linv, MMatrix<Complex>  &Ginv,  MMatrix<Complex> &projA,const int NS,const int K,Data_in & Data,ResultsStore &additResults,const Complex Psi,VVector<Complex> &SX, VVector<Complex> &deltaSX, VVector<Complex> &PdS,double put_to_memory)
{

  if(put_to_memory==1 || put_to_memory==0)
  {
    LinvS(X,Linv,Data,additResults,NS,K,SX);  //useful ???
    LinvdeltaS(X,Linv,deltaSX,Data,additResults,Psi,NS,put_to_memory); //not optimal

    Complex aux_num;
    int i,j;
    for(i=0; i<NS; i++)
    {
      aux_num=0;
      for(j=0; j<NS; j++) aux_num+=SX[j]*projA[i][j];
      PdS[i] = deltaSX[i] - aux_num;
    } //we could define Id(NS,NS) and do multiply(Id-ProjA,SX)
  }
  if(put_to_memory==2)
  {
    Complex aux_num;
    int i,j;
   // show(X.deltaSX);
   // cout <<"\n ----------";
    for(i=0; i<NS; i++)
    {
      aux_num=0;
      for(j=0; j<NS; j++) aux_num+=X.SX[j]*projA[i][j];
      PdS[i] = X.deltaSX[i] - aux_num;
    }
   // show(PdS);getchar();
  }
  //  return PdS;
}

// Once we have the correct deformations PdeltaLinvS we can compute the final metric

double gpsipsiF(vector<Point> &X, MMatrix<Complex> &Linv, MMatrix<Complex> &Ginv, MMatrix<Complex> &projA,Data_in & Data,ResultsStore &additResults,const int NS,const int Number_Points,const int K, const Complex Psi,double put_to_memory)
{

  double gppbar;
  Complex gpp=0;
  Complex D_function;
  VVector<Complex> SX,deltaSX,PdS;
  SX.resize(NS,1);
  deltaSX.resize(NS,1);
  PdS.resize(NS,1);
  int i;
  for(int n_points=0; n_points<Number_Points; n_points++)
  {
    if ( n_points % ( ( int ) Number_Points/100 ) ==0 )
    {
      cout.flush();
      cout << "*** Computing the WP metric  "<< ( 100.*n_points ) / ( Number_Points ) << "%             \r" ;
    }
    PdeltaLinvS(X[n_points], Linv, Ginv, projA,NS,K, Data,additResults, Psi,SX,deltaSX, PdS,put_to_memory);//the projection of deltaSX
    D_function=X[n_points].D_function;
    for(i=0; i<NS; i++)
    {
      gpp += pow2(abs(PdS[i]))*X[n_points].mass/D_function; // We cannot use Ginv: We are in the orthonormal basis already!!!!!
    }
  }

  gppbar = abs(gpp)*(1./(Number_Points*pow(K,3.))); //TODO different from Sergio's code   *N/Volume;
  return gppbar;
}
