/***************************************************************************
 *   Copyright (C) 2007 by Julien Keller   *
 *   j.keller@imperial.ac.uk   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef CALABIYAU_H 
#define CALABIYAU_H
// -llapack -lboost_thread -O3  -pipe -mno-sse2 -ffast-math  -funroll-all-loops  for options of compilation
///////////////////////
//Standard Libraries //
///////////////////////FDx_isA
#include <stdio.h> //standard basic library
#include <stdlib.h> //standard basic library
#include <cassert> //standard library for assert --not sure it is necessary here
#include <iostream> //standard library for text input and output
#include <string>//standard library for easy to use strings
#include <sstream> //standard library to make combining strings easier
#include <vector> //standard library for the std::vector class
#include <fstream>
#include <iomanip>
#include <sys/time.h> //get the time to check speed

///////////////////////////////////////////////
// Contains global variables like the dimension 
///////////////////////////////////////////////
//TODO should be given as a parameter
#include "NPFile.h"
//////////////////////
//Output libraries ///
//////////////////////
#include "in-output.h" //takes care of input and output

//////////////////////
//Matrices packages //
/////////////////////

#include "MMatrix.h" //library to handle matrices in memory quickly
#include "m_matrix2.h" //library for standard procedures on matrices

/////////////////////
// GENERATE POINTS //
/////////////////////
#include "GeneratePoints.h"

 
//#include "Symmetries.h"
//#include "Sections.h"
/////////////////////////////////////////
//Methods to compute the balanced metrics
/////////////////////////////////////////
#include "methods.h"

////////////////////////////////////////
// Computation of the curvatures
///////////////////////////////////////
#include "Curvatures.h"

//////////////////////////////////////////////
// Computation of the Weil Petersson metric //
//////////////////////////////////////////////
#include "WP_procedures.h"

//Mathematical Constants and Macros
#define Complex  std::complex<double>
#define _Complex_I std::complex<double>(0, 1)
#define _1  (std::complex<double>(1, 0))
#define _2  (std::complex<double>(2, 0))
#define _0  (std::complex<double>(0, 0))
#define PI  3.14159265358979323846

#define pow2(x) (x*x)
#define pow3(x) (x*x*x)
#define pow4(x) pow2(x*x)
#define pow5(x) (x*pow4(x))
#define pow6(x) pow2(pow3(x))
#define pow8(x) pow2(pow4(x))

/////////////////////////////////////////// //////////////////
// Important variables to run the program -- Default choices
//////////////////////////////////////////////////////////////
int Number_Points= 3000; //which means we compute exactly NI*Number_Points
 //for dual core CPUS :
#define NUMBER_CPUS 2  
int Number_Iterations= 2 * NUMBER_CPUS; // number of iterations for computing points (should be 2 for dual core cpus)
Complex Psi_start=(std::complex<double>(0, 0.1)); // cpx structure parameter in equation \sum_i x_i^{n+1} -\Psi\Pi_i x_i = 0
double radius_Psi_end=0;
double angle_Psi_end=0; // cpx structure parameter in equation \sum_i x_i^{n+1} -\Psi\Pi_i x_i = 0
double radius_Psi_step=0.1;
double angle_Psi_step=0.1;

double  ERR=1.0e-10;
int precision_generate_points=5;
int precision_output=5;
int K=5; // this is the degree of the line bundle
int Einstein=0;
int read_points_from_file=0;
int Refined=0;
double print_ratio=0;//VolumeFormFSmetric/Omega
double ratio_refined=100;  // % of points computed for refined version
double print_scalar=1; //print or not the scalar curv of the points
double print_coeff=1; //prints scalar curvature pointwisely
double refined_nb_iterations=2;
double put_to_memory=0;
#endif


