/***************************************************************************
 *   Copyright (C) 2007 by Julien Keller   *
 *   j.keller@imperial.ac.uk   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef IPOINT
#define IPOINT

 
#include "in-output.h"
#include "NPFile.h"
#include "m_matrix2.h"
#define Complex  std::complex<double>



class Point{
  public:
    VVector<Complex> comp;
    double mass;
    MMatrix<Complex> Result_sections; //will contain evaluation of sections at the point (size = Degree*(K+1))
    Complex affine[Degree]; //Affixe coordinates 
    double D_function;
    VVector<Complex> SX; //values of sections at the point 
    VVector<Complex> deltaSX; //values of delta sections at the point
    vector<vector<int> > Se;
    int length() const { return(comp.getHeight());}   
   //constructors - -destructors
  //  ~Point(){ ~comp;} 
    Point() {};
    Point(int size_point){comp.resize(size_point,1);}
    void resize(int l){comp.resize(l,1);}
    //TODO should be dependant only on the size of w
    Point(Complex w[]) {for(int i=0; i<comp.getHeight(); i++) comp[i]=w[i]; mass=0;}
    Point(Complex w[], double m) {for(int i=0; i<comp.getHeight(); i++) comp[i]=w[i]; mass=m;}
    //functions
    void print()
    {
      show(comp);
    } 
 /*   void print(char* a) TODO
    {
      show(comp,a);
    }*/
    void scalar(const Complex t){
      comp*=t;
    }
    
    void toprojective(){
      int i,j,position=0;
      double max_element=abs(comp[0]);
      comp.resize(comp.getHeight()-1,1); 
      for(i=1;i<comp.getHeight();i++){
        if(abs(comp[i])>max_element){
          max_element=abs(comp[i]);
          position=i;
        }
      }
      j=0;
      for(i=0;i<comp.getHeight();i++){
        if(i!=position){
          comp[j]=comp[i]/comp[position];
          j++;
        }
      }
    }
     
    Point scalar_(const Complex t){
      Point Pnew(comp.getHeight());//TODO :one should be able to do that straightforward using MMAtrix package.
      for(int i=0;i<comp.getHeight();i++){
        Pnew.comp[i]=t*comp[i];
      }
      Pnew.mass=mass;
      return(Pnew);
    }
    Point toprojective_(){
      int i,j,position=0;
      double max_element=abs(comp[0]);
      Point Pnew(comp.getHeight()-1); 
      for(i=1;i<comp.getHeight();i++){
        if(abs(comp[i])>max_element){
          max_element=abs(comp[i]);
          position=i;
        }
      }
      j=0;
      for(i=0;i<comp.getHeight();i++){
        if(i!=position){
          Pnew.comp[j]=comp[i]/comp[position];
          j++;
        }
      }
      Pnew.mass=mass;
      return(Pnew);
     
    }

};

//typedef vector<Complex> Point; DOES NOT work, Complex might not have some constructor
/*
class Point  // one pt on P^D=C^{D+1=N}
{
  public:
    Complex comp[N];
    double mass;
    Point() {};
    Point(Complex w[]) {for(int i=0; i<N; i++) comp[i]=w[i]; mass=0;}
    Point(Complex w[], double m) {for(int i=0; i<N; i++) comp[i]=w[i]; mass=m;}
    void print() {for(int i=0; i<N; i++) {cout << comp[i] << ' ';} cout << mass << '\n';}
};
*/

template<class T> T read(ifstream &file) {T x(N); for(int i=0; i<N; i++) file >> x.comp[i]; file >> x.mass; return x;}

template<class T>
    void write(const T &x, ofstream &file) {for(int i=0; i<N; i++) file << x.comp[i] ; file << x.mass << '\n';}

struct sheaf
{           // useful for threading
  int np, st;
  sheaf(int i, int j) {np=i; st=j;}
  void print() {cout<< np <<' '<< st <<'\n';}
};

void Eta(const VVector<double> &v);
void EtaV(const vector<VVector<double> > &v);
///void errors(const vector<Matrix> &B, Matrix &Mean, MMatrix<double> &StdDev);

VVector<double> real(const VVector<Complex> &v);
//VectorD abs(const Vector &v);
MMatrix<double> abs(const MMatrix<Complex> &G);


#endif

