/***************************************************************************
 *   Copyright (C) 2007 by Julien Keller                                   *
 *   jkeller@imperial.ac.uk                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "methods.h"

#define PI  3.14159265358979323846
#define E_  2.71828182845904523536
#define pow2(x) (x*x)
//This is from this file that we compute the FSoHilb map.
//Several procedures do exactly the same thing but are written using different structures
//for instance , use different matrices package.



double launch_method_pointwise_1a ( const int Number_Points,const double ERR,const int Einstein,const int step,const Complex Psi,int method_refined,const int K,Data_in & Data, double error_prev  )
{
  int i;
  int  NS=Data.Sec.size();

  double denom;

  Data.ValuePoints.fillWith ( 1 ); //in this way the first row only contains the value 1 for all iterations, i.e all points
  if ( ( Einstein==1 ||Einstein==2 ) && step==0 )
  {
    Data.Extra_Masses.resize( Number_Points,1 );
    //this contains the denom at the previous iteration.
  }

  //definition of matrices here to save time in memory
  MMatrix<Complex> df ( Degree,1 ),ddf ( Degree,Degree ),DxisA ( NS,Dimension ),sAy ( NS,1 ),
  DxP ( Dimension,1 ),DbD2xP ( Dimension,Dimension ),g ( Dimension,Dimension ),FS ( Degree,Degree ), O ( Dimension,Dimension ),v_deriv ( Dimension,Degree );
  vector<vector<int> > Se;
  ////////////////////////////////////////////////////
  ResultsStore additResults;
  additResults.store_Results ( Data.NS,K,Data.Sec,Data.Bs0,Data.Bl0,Data.Bs,Data.Pairs,Data.vI_new2 );
  //TODO use systemetically this structure
  ///it is that part that does all the computations and takes
  ///most of the time.
  Data.v_new2.fillWith ( 0 );
  double detg,ratio;
  Data.vI_new_horner=coeffsToHorner ( Data.vI_new2 );
  ///returns 1 and a_{i}/a_{i+1} of vI_new2
  MMatrix<Complex> prod ( Data.BS,1 );
  MMatrix<double> Results(6,1);
  MMatrix<Complex> InversionTemp(Data.G.getHeight(),Data.G.getHeight());
  MMatrix<Complex> InversionTempsquare(Data.G.getHeight(),Data.G.getHeight());
  MMatrix<Complex> G_temp(Data.G.getHeight(),Data.G.getHeight());
  MMatrix<Complex> Identity(Data.G.getHeight(),Data.G.getHeight());
  Identity.fillWith(0);
  for(int l=0;l<Identity.getHeight();l++)
  {
    Identity[l][l]=1.;
  }
  Results.fillWith(1);
  Results[3][0]=0;Results[4][0]=0;
  Results[0][0]=0;
  Results[5][0]=0;
  VVector<double> PointsDet;
  double detgtemp;
  double partial_mass_new=0;
  double partial_mass=0;
  //  Data.ratio_refined=100;
  if ( method_refined==1)
  {
    PointsDet.resize(((int) ((Data.ratio_refined/100.)*Number_Points+1))+1,1);
    for(i=0;i/((double) Number_Points)<(Data.ratio_refined/100.);i++)
    {
      PointsDet[i]=VolumeForm_FSmetric2_FAST2 ( Data.X [i ],additResults,Data.G,df,ddf,DxisA,Se,sAy,DxP,DbD2xP,g,FS,O,v_deriv,Psi );

      partial_mass+=Data.X[i].mass;
      partial_mass_new+= PointsDet[i];
      if ( i % ( ( int ) (Number_Points/100) ) ==0 )
      {
        cout.flush();
        cout << "  T-map computation (refined method, 1rst step): " << (int) ((100./Data.ratio_refined)*( 100*i ) / ( Number_Points )) << "%  \r" ;
      }
    }
    for(i=0;i/((double) Number_Points)<(Data.ratio_refined/100.);i++)
    {
      PointsDet[i]*=(partial_mass)/partial_mass_new; // we need to normalize the masses in order to solve the calabi problem.
    }
    //cout << "\n Masse  ------>" << partial_mass_new/Number_Points <<  "   " <<partial_mass/Number_Points<< "ratio "<< partial_mass_new/partial_mass << "\n";
  }
  prod.fillWith ( 1 );
  for ( int p=0;p<Data.BS;p++ )
  {
    for ( int q=0;q<=p;q++ )
    {
      prod[p][0]*=Data.vI_new_horner[q][0]; //returns a_p
    }
  }

  for ( i=0; i<Number_Points; i++ )
  {
  //   cout << "  *** "<<DerivY4(Data.X[i],Psi)<< "  "<< DerivY4bis(Data.X[i],Psi); getchar();
    if ( i % ( ( int ) Number_Points/100 ) ==0 )
    {
      cout.flush();
      if (method_refined==0)
        cout << "  T-map computation: " << ( 100.*i ) / ( Number_Points ) << "%  \r" ;
      if (method_refined==1)
        cout << "  T-map computation (refined method, 2nd step ) :  " << ( 100.*i ) / ( Number_Points ) << "%  \r" ;
    }

    sAlpha_pointwise2a ( K,Data,Data.X [ i] );
    
    /*VVector<Complex> S2; 
    S2.resize(NS,1);
    cout <<"  TEST " << Data.sApointwise2.getHeight() << endl;
    getchar();
    show(Data.sApointwise2);
        PutToFullVector(Data.sApointwise2,S2,Data.PairsL,Data.Bs,Data.Bl);
   show(S2); 
     getchar();*/  
    denom=DsAlphaBeta_pointwise2a ( Data );


  /*  MMatrix<Complex> compteurs(Data.G.getHeight(),1);
    int ic;
    compteurs.fillWith(0);
    compteurs[0][0]=i;  // current point x
    for(ic=1;ic<Data.G.getHeight(),ic++)
    {
        for(compteurs[ic][0]=compteurs[ic-1][0]; compteurs[ic][0]<Number_Points; compteurs[ic][0]++)
        {
            
           

*/


    if ( method_refined==0 && Einstein==0 )
    {
      Data.v_new2+= ( Data.X [ i ].mass/denom ) *Data.sABpointwise2;
      for ( int p=0;p<Data.BS;p++ )
      {
        Data.vI_new_old_ratio[p][0]+= ( Data.X [i ].mass/denom ) *prod[p][0]*Data.sABpointwise2[p][0]; //TODO should be possible with just one line where is it used ???
      }

    }

    if (method_refined==0 && Einstein==1 ) //this time we change the integration slightly
    {
      if ( step==0 )
      {

        Data.v_new2+= ( Data.X [ i ].mass/denom ) *Data.sABpointwise2;
      }
      else
      {
        Data.v_new2+= ( Data.X [ i ].mass/denom ) *Data.sABpointwise2*pow ( exp ( Data.Extra_Masses[i][0]-denom ),1./K );
        //   cout << "             --- > " <<Extra_Masses[i][0]-denom;

      }
      Data.Extra_Masses[i][0]=denom;
    }
    if ( method_refined==1 )
    {

      if(i/((double) Number_Points)<(Data.ratio_refined/100.))
      {
        ratio=PointsDet[i]/Data.X[i].mass;
        Data.v_new2+=( Data.X [ i ].mass/denom )*( ratio-1 ) *Data.sABpointwise2;//later we divise by the number of points
      }
      else
      {
        ratio=1;
      }
      if(Data.print_ratio==1)
      {
        cout<<"\n-->ratio VolumeFormFSmetric/Omega (should be close to 1)="<<ratio ;
        ///    getchar();
      }
      if(ratio>Results[0][0])
      {
        Results[0][0]=ratio;//max
      }
      if(ratio<Results[2][0])
      {
        Results[2][0]=ratio;//min
      }
      Results[1][0]=(Results[1][0]*i+ratio)/(i+1);  //simple average
      Results[3][0]+=ratio*Data.X[i].mass/Number_Points;//average of ratio
      Results[4][0]+=abs(1-ratio)*Data.X[i].mass/Number_Points; //L^1 norm of ratio-1

    }

  }
  if ( method_refined==1 )
  {
    cout<<"--> Max/Average/Min for ratio Volume_Form_FS_metric/Volume_Form : \n \033[31m" << Results[0][0] << "/" << Results[1][0] << "/" << Results[2][0];
    cout<<"\033[0m";
    cout<<"\n-->  L^1 norm of Error : \033[31m" << Results[4][0]/(Results[3][0]); 
         //<< " "<< Results[3][0] ; //L^1 norm of |ratio -1|.
    cout<<"\033[0m";

  }
  MMatrix<Complex> vI_new2_save;
  if ( method_refined==0 )
  {
    cout << "\n--> Bergman function ="<< Data.v_new2.sum() << "\n--> Error [in computation of Sections]=" << abs ( imag ( Data.v_new2.sum() )*100 /(real ( Data.v_new2.sum() )) ) << " %"<< endl;
    //the sum can ve very large
     if(no_use_symmetry==1)
     {
     //  cout << "size = "<< Data.v_new2.getHeight() << "  " << Data.v_new2.getWidth();
       show(Data.v_new2);
       getchar();
       PutToFullMatrix  ( Data.v_new2 ,Data.G_new,Data.PairsL,Data.Bs,Data.Bl );

     }
     else{
     PutToFullMatrix  ( average2 (  ( Data.v_new2 ),Data.Bs0,Data.Bl0 ),Data.G_new,Data.PairsL,Data.Bs,Data.Bl );
     }
    cout<< "\n\033[34m**  Inverting \033[0m \n";
    ///   show(Data.G_new);
    //getchar();
    Invert2( Data.G_new,Data.Gi_new,ERR ); // we inverse the metric here
    ///Invert( Data.G_new,Data.Gi_new,ERR ); // we inverse the metric here using a less efficient library blas ?
    cout<< "--> Inversion done. \n";
    vI_new2_save=Data.vI_new2; //just to know the evolutions between 2 iterations
    ///cout << "\n inverse "<< Data.G_new;
     if(no_use_symmetry==1)
     {
       for(int k=0;k<Data.vI_new2.getHeight();k++)
       {
         Data.vI_new2[k][0]=Data.Gi_new[Data.Pairs[k][0]][Data.Pairs[k][1]];

       }
     }
     else{
      GetFromMMatrix( Data.vI_new2, Data.Gi_new, Data.Pairs, Data.Bs0,Data.Bl0 );
     }
  }
  if ( method_refined==1 ) //here we do not need to inverse , this is a kind of Newton method close to the point we are looking for.
  {
    Data.v_new2*= ( 1./(Number_Points ));
    PutToFullMatrix(average2 ((Data.v_new2 ),Data.Bs0,Data.Bl0),Data.Eta,Data.PairsL,Data.Bs,Data.Bl );
    //  show(Data.Eta);

    /// Data.G+=(E_)*multiply ( multiply ( Data.G,Data.Eta ),Data.G );  //exp(1) should be the best constant
    double constant=E_*E_; //from Liu-Ma 's paper.
    InversionTemp=  constant*( multiply ( Data.G,Data.Eta ) ); ///TODO should  be 2E_ or more: not clear what is the best constant to choose
    InversionTempsquare = Identity;

    for(int k=0;k<3;k++)
    {
      G_temp+=InversionTempsquare;
      InversionTempsquare=multiply(InversionTempsquare,InversionTemp);
    }

    Data.G=multiply(G_temp,Data.G);
    vI_new2_save=Data.vI_new2; //just to know the evolutions between 2 iterations
    // PutToVector ( Data.vI_new2, Data.G, Data.Pairs, Data.Bs0,Data.Bl0 );
    GetFromMMatrix( Data.vI_new2, Data.G, Data.Pairs, Data.Bs0,Data.Bl0 );

  }

  Data.vI_new2*=1./Data.vI_new2[0][0]; //just normalization
  double return_error=( Data.vI_new2-vI_new2_save ).norm_euclid();
  /* if(method_refined==1 && abs(Results[1][0]-1)<0.01) /// 0.01 ?? could be changed as a parameter
   {
     return_error=2*error_prev; //it makes it stop
   }
   if(method_refined==1 && return_error>2*error_prev)
   {
     Data.vI_new2=vI_new2_save;
   }
  */
  return ( return_error );

}


void global_launch_method_pointwise_1a ( const int Number_Iterations,int Einstein,const int Number_Points,const int K,const double ERR,const Complex Psi,const int Refined, const double refined_nb_iterations, const double print_coeff, Data_in & Data )
{

  int j;
  Data.sApointwise2.resize ( Data.NS,1 );
  Data.v_new2.resize ( Data.Pairs.size(),1 );
  Data.sABpointwise2.resize ( Data.Pairs.size(),1 );
  Data.ValuePoints.resize ( degree_polynomial_f(),K+1 );  //this is a very small thing around = 6*15*complex. should be MMatrix
  Data.G_new.resize ( Data.NS,Data.NS );
  Data.Gi_new.resize ( Data.NS,Data.NS );
  Data.G.resize ( Data.NS,Data.NS ); //this contains the hermitian matrice
  //  MMatrix<double> Eta;
  //  MMatrix<double> Extra_Masses;
  double error=1e12,error_prev;
  int jmax= ((int) Number_Iterations);

  int method_refined=0; //in order to know when we start the refined method

  for ( j=0; j<jmax; j++ )
  {
    error_prev=error;
    error=launch_method_pointwise_1a ( Number_Points,ERR,Einstein,j,Psi,method_refined, K,Data,error_prev );
    cout << "\n--> Difference between the two computed metrics (step "<<j+1<<") :\033[31m"<< error<<"\033[0m";
    ///gives the norm of the difference of the two hermitian metrics at step j
 
    if(method_refined==0 && Refined==1 && error <ERR ) // we can launch the refined method because we are not too far from a balanced point
    {
      error=1e12;
      j=-1; //we start again at 0 and stop at refined_nb_iterations-1
      jmax=(int) refined_nb_iterations;
      error_prev=error;
      method_refined=1;
      Einstein=-1;
      Data.Eta.resize ( Data.NS,Data.NS );
      cout << "\n \n -------------------------------------------------------------";
      cout << "\n\033[34m ** Starting refined method with % computed points =\033[0m"<<Data.ratio_refined;
      cout << " \n -------------------------------------------------------------";
    }
    if(Refined==0 &&  error<ERR)
    {
      cout <<"\n *** Stopping normal method : "<< error << " " << error_prev <<endl;
      cout << "\nResult at step " <<j+1<< " " <<endl; 
      showtranspose ( (Data.vI_new2) );
       j=jmax+1; //stop the process;
      
    }

    if(method_refined==1 &&  error<ERR)
      //method_refined==1 && (error>2*error_prev || error<ERR))
    {
      cout <<"\n *** Stopping refined method : "<< error << " " << error_prev <<endl;
      cout << "\nResult at step " <<j+1<< " " <<endl; 
      showtranspose ( (Data.vI_new2) );
       j=jmax+1; //stop the process;
      
    }

    //  launch_method_pointwise_0(Number_Points,Pairs,Sec,BS,X,K,Bl0,Bl,Bs0,Bs,PairsL,ERR,vI_new);
    // This is for the pointwise algo
    if(print_coeff==1 && j!=jmax+1){
    cout << "\nResult at step " <<j+1<< " " <<endl; 
    showtranspose ( (Data.vI_new2) );
    // printV(vI_new);
     
        }
    printf ( "\n" );

  }
}







